import os
from datetime import datetime
from pathlib import Path
from unittest import result

from click.testing import CliRunner
from feathr import (BOOLEAN, FLOAT, INT32, FeatureQuery, ObservationSettings,
                    SparkExecutionConfiguration, TypedKey, ValueType)
from feathr.client import FeathrClient
from feathr.utils.job_utils import get_result_df

from test_fixture import basic_test_setup
from feathr.constants import OUTPUT_FORMAT

# test parquet file read/write without an extension name
def test_feathr_get_secrets_from_key_vault():
    """
    Test if the program can read the key vault secrets as expected
    """
    # TODO: need to test get_environment_variable() as well
    os.environ['SECRETS__AZURE_KEY_VAULT__NAME'] = 'feathrazuretest3-kv'

    # the config below doesn't have `ONLINE_STORE__REDIS__HOST` for testing purpose
    yaml_config = """
    project_config:
        project_name: 'project_feathr_integration_test'
    offline_store:
        s3:
            s3_enabled: true
            s3_endpoint: 's3.amazonaws.com'
        snowflake:
            url: "dqllago-ol19457.snowflakecomputing.com"
            user: "feathrintegration"
            role: "ACCOUNTADMIN"
    spark_config:
        spark_cluster: 'databricks'
        spark_result_output_parts: '1'
        azure_synapse:
            dev_url: 'https://feathrazuretest3synapse.dev.azuresynapse.net'
            pool_name: 'spark3'
            workspace_dir: 'abfss://feathrazuretest3fs@feathrazuretest3storage.dfs.core.windows.net/feathr_test_workspace'
            executor_size: 'Small'
            executor_num: 1

        databricks:
            workspace_instance_url: 'https://adb-2474129336842816.16.azuredatabricks.net/'
            workspace_token_value: ''
            config_template: '{"run_name":"FEATHR_FILL_IN","new_cluster":{"spark_version":"9.1.x-scala2.12","num_workers":1,"spark_conf":{"FEATHR_FILL_IN":"FEATHR_FILL_IN"},"instance_pool_id":"0403-214809-inlet434-pool-l9dj3kwz"},"libraries":[{"jar":"FEATHR_FILL_IN"}],"spark_jar_task":{"main_class_name":"FEATHR_FILL_IN","parameters":["FEATHR_FILL_IN"]}}'
            work_dir: 'dbfs:/feathr_getting_started'
            feathr_runtime_location: ''
    online_store:
        redis:
            port: 6380
            ssl_enabled: True
    feature_registry:
        purview:
            type_system_initialization: false
            purview_name: 'feathrazuretest3-purview1'
            delimiter: '__'
    """

    with open("/tmp/feathr_config.yaml", "w") as text_file:
        text_file.write(yaml_config)

    client = FeathrClient(config_path="/tmp/feathr_config.yaml")
    # `redis_host` should be there since it's not available in the environment variable, and not in the config file, we expect we get it from azure key_vault
    assert client.redis_host is not None




from typing import Any, Dict, List, Optional, Tuple


class SparkExecutionConfiguration:
    """A wrapper class to enable Spark Execution Configurations which will be passed to the underlying spark engine.
    Attributes:
        spark_execution_configuration: dict[str, str] which will be passed to the underlying spark engine
    Returns:
        dict[str, str]
    """
    def __new__(cls, spark_execution_configuration = Dict[str, str]) -> Dict[str, str]:
        return spark_execution_configuration



import importlib
import inspect
import json
import logging
from loguru import logger
import os
from pathlib import Path
import sys
from urllib.parse import urlparse
from uuid import UUID
from azure.identity import DefaultAzureCredential
from typing import Any, Dict, List, Optional, Tuple
from re import sub
from click import option
from jinja2 import Template

import requests
from feathr.constants import INPUT_CONTEXT, TYPEDEF_ANCHOR, TYPEDEF_ANCHOR_FEATURE, TYPEDEF_DERIVED_FEATURE, TYPEDEF_SOURCE
from feathr.definition.anchor import FeatureAnchor
from feathr.definition.dtype import FeatureType, str_to_value_type, value_type_to_str
from feathr.definition.feature import Feature, FeatureBase
from feathr.definition.feature_derivations import DerivedFeature
from feathr.definition.repo_definitions import RepoDefinitions
from feathr.definition.source import GenericSource, HdfsSource, InputContext, JdbcSource, Source
from feathr.definition.transformation import ExpressionTransformation, Transformation, WindowAggTransformation
from feathr.definition.typed_key import TypedKey
from feathr.registry.feature_registry import FeathrRegistry
from feathr.utils._file_utils import write_to_file

def to_camel(s):
    if not s:
        return s
    if isinstance(s, str):
        if "_" in s:
            s = sub(r"(_)+", " ", s).title().replace(" ", "")
            return ''.join([s[0].lower(), s[1:]])
        return s
    elif isinstance(s, list):
        return [to_camel(i) for i in s]
    elif isinstance(s, dict):
        return dict([(to_camel(k), s[k]) for k in s])


def _topological_sort(derived_feature_list: List[DerivedFeature]) -> List[DerivedFeature]:
    """
    In the current registry implementation, we need to make sure all upstream are registered before registering one derived feature
    Thus we need to sort derived features by the partial order of dependencies, upstream to downstream.
    """
    ret = []
    # We don't want to destroy the input list
    input = derived_feature_list.copy()

    # Each round add the most downstream features into `ret`, so `ret` is in reversed order
    while input:
        # Process all remaining features
        current = input.copy()

        # In Python you should not alter content while iterating
        current_copy = current.copy()

        # Go over all remaining features to see if some feature depends on others
        for f in current_copy:
            for i in f.input_features:
                if i in current:
                    # Someone depends on feature `i`, so `i` is **not** the most downstream
                    current.remove(i)

        # Now `current` contains only the most downstream features in this round
        ret.extend(current)

        # Remove one level of dependency from input
        for f in current:
            input.remove(f)

    # The ret was in a reversed order when it's generated
    ret.reverse()

    if len(set(ret)) != len(set(derived_feature_list)):
        raise ValueError("Cyclic dependency detected")
    return ret


class _FeatureRegistry(FeathrRegistry):
    def __init__(self, project_name: str, endpoint: str, project_tags: Dict[str, str] = None, credential=None, config_path=None):
        self.project_name = project_name
        self.project_tags = project_tags
        self.endpoint = endpoint
        self.credential = DefaultAzureCredential(
            exclude_interactive_browser_credential=False) if credential is None else credential
        self.project_id = None

    def register_features(self, workspace_path: Optional[Path] = None, from_context: bool = True, anchor_list=[], derived_feature_list=[]):
        """Register Features for the specified workspace. 
        Args:
            workspace_path (str, optional): path to a workspace. Defaults to None, not used in this implementation.
            from_context: whether the feature is from context (i.e. end users has to callFeathrClient.build_features()) or the feature is from a pre-built config file. Currently Feathr only supports register features from context.
            anchor_list: The anchor list after feature build
            derived_feature_list: the derived feature list after feature build
        """
        if not from_context:
            raise RuntimeError(
                "Currently Feathr only supports registering features from context (i.e. you must call FeathrClient.build_features() before calling this function).")

        # Before starting, create the project
        self.project_id = self._create_project()

        for anchor in anchor_list:
            source = anchor.source
            # 1. Create Source on the registry
            # We always re-create INPUT_CONTEXT as lots of existing codes reuse the singleton in different projects
            if (source.name == INPUT_CONTEXT) or (not hasattr(source, "_registry_id")):
                source._registry_id = self._create_source(source)
            # 2. Create Anchor on the registry
            if not hasattr(anchor, "_registry_id"):
                anchor._registry_id = self._create_anchor(anchor)
            # 3. Create all features on the registry
            for feature in anchor.features:
                if not hasattr(feature, "_registry_id"):
                    feature._registry_id = self._create_anchor_feature(
                        anchor._registry_id, feature)
        # 4. Create all derived features on the registry
        for df in _topological_sort(derived_feature_list):
            if not hasattr(df, "_registry_id"):
                df._registry_id = self._create_derived_feature(df)
        url = '/'.join(self.endpoint.split('/')[:3])       
        logger.info(f"Check project lineage by this link: {url}/projects/{self.project_name}/lineage")

    def list_registered_features(self, project_name: str) -> List[str]:
        """List all the already registered features. If project_name is not provided or is None, it will return all
        the registered features; otherwise it will only return features under this project
        """
        resp = self._get(f"/projects/{project_name}/features")
        # In V1 API resp should be an array, will be changed in V2 API
        return [{
            "name": r["attributes"]["name"],
            "id": r["guid"],
            "qualifiedName": r["attributes"]["qualifiedName"],
        } for r in resp]

    def get_features_from_registry(self, project_name: str) -> Tuple[List[FeatureAnchor], List[DerivedFeature]]:
        """
        [Sync Features from registry to local workspace, given a project_name, will write project's features from registry to to user's local workspace]

        Args:
            project_name (str): project name.
        """
        lineage = self._get(f"/projects/{project_name}")
        return dict_to_project(lineage)

    def _create_project(self) -> UUID:
        r = self._post(f"/projects", {"name": self.project_name})
        self.project_id = UUID(r["guid"])
        return self.project_id

    def _create_source(self, s: Source) -> UUID:
        r = self._post(
            f"/projects/{self.project_id}/datasources", source_to_def(s))
        id = UUID(r["guid"])
        s._registry_id = id
        s._qualified_name = f"{self.project_name}__{s.name}"
        return id

    def _create_anchor(self, s: FeatureAnchor) -> UUID:
        r = self._post(
            f"/projects/{self.project_id}/anchors", anchor_to_def(s))
        id = UUID(r["guid"])
        s._registry_id = id
        s._qualified_name = f"{self.project_name}__{s.name}"
        return id

    def _create_anchor_feature(self, anchor_name: str, s: Feature) -> UUID:
        r = self._post(
            f"/projects/{self.project_id}/anchors/{anchor_name}/features", feature_to_def(s))
        id = UUID(r["guid"])
        s._registry_id = id
        s._qualified_name = f"{self.project_name}__{anchor_name}__{s.name}"
        return id

    def _create_derived_feature(self, s: DerivedFeature) -> UUID:
        r = self._post(
            f"/projects/{self.project_id}/derivedfeatures", derived_feature_to_def(s))
        id = UUID(r["guid"])
        s._registry_id = id
        s._qualified_name = f"{self.project_name}__{s.name}"
        return id

    def _get(self, path: str) -> dict:
        logging.debug("PATH: ", path)
        return check(requests.get(f"{self.endpoint}{path}", headers=self._get_auth_header())).json()

    def _post(self, path: str, body: dict) -> dict:
        logging.debug("PATH: ", path)
        logging.debug("BODY: ", json.dumps(body, indent=2))
        return check(requests.post(f"{self.endpoint}{path}", headers=self._get_auth_header(), json=body)).json()

    def _get_auth_header(self) -> dict:
        return {"Authorization": f'Bearer {self.credential.get_token("https://management.azure.com/.default").token}'}
    
    @classmethod
    def _get_py_files(self, path: Path) -> List[Path]:
        """Get all Python files under path recursively, excluding __init__.py"""
        py_files = []
        for item in path.glob('**/*.py'):
            if "__init__.py" != item.name:
                py_files.append(item)
        return py_files

    @classmethod
    def _convert_to_module_path(self, path: Path, workspace_path: Path) -> str:
        """Convert a Python file path to its module path so that we can import it later"""
        prefix = os.path.commonprefix(
            [path.resolve(), workspace_path.resolve()])
        resolved_path = str(path.resolve())
        module_path = resolved_path[len(prefix): -len(".py")]
        # Convert features under nested folder to module name
        # e.g. /path/to/pyfile will become path.to.pyfile
        return (
            module_path
            .lstrip('/')
            .replace("/", ".")
        )

    @classmethod
    def _extract_features_from_context(self, anchor_list, derived_feature_list, result_path: Path) -> RepoDefinitions:
        """Collect feature definitions from the context instead of python files"""
        definitions = RepoDefinitions(
            sources=set(),
            features=set(),
            transformations=set(),
            feature_anchors=set(),
            derived_features=set()
        )
        for derived_feature in derived_feature_list:
            if isinstance(derived_feature, DerivedFeature):
                definitions.derived_features.add(derived_feature)
                definitions.transformations.add(
                    vars(derived_feature)["transform"])
            else:
                raise RuntimeError(
                    "Object cannot be parsed. `derived_feature_list` should be a list of `DerivedFeature`.")

        for anchor in anchor_list:
            # obj is `FeatureAnchor`
            definitions.feature_anchors.add(anchor)
            # add the source section of this `FeatureAnchor` object
            definitions.sources.add(vars(anchor)['source'])
            for feature in vars(anchor)['features']:
                # get the transformation object from `Feature` or `DerivedFeature`
                if isinstance(feature, Feature):
                    # feature is of type `Feature`
                    definitions.features.add(feature)
                    definitions.transformations.add(vars(feature)["transform"])
                else:
                    raise RuntimeError("Object cannot be parsed.")

        return definitions

    @classmethod
    def _extract_features(self, workspace_path: Path) -> RepoDefinitions:
        """Collect feature definitions from the python file, convert them into feature config and save them locally"""
        os.chdir(workspace_path)
        # Add workspace path to system path so that we can load features defined in Python via import_module
        sys.path.append(str(workspace_path))
        definitions = RepoDefinitions(
            sources=set(),
            features=set(),
            transformations=set(),
            feature_anchors=set(),
            derived_features=set()
        )
        for py_file in self._get_py_files(workspace_path):
            module_path = self._convert_to_module_path(py_file, workspace_path)
            module = importlib.import_module(module_path)
            for attr_name in dir(module):
                obj = getattr(module, attr_name)
                if isinstance(obj, Source):
                    definitions.sources.add(obj)
                elif isinstance(obj, Feature):
                    definitions.features.add(obj)
                elif isinstance(obj, DerivedFeature):
                    definitions.derived_features.add(obj)
                elif isinstance(obj, FeatureAnchor):
                    definitions.feature_anchors.add(obj)
                elif isinstance(obj, Transformation):
                    definitions.transformations.add(obj)
        return definitions

    @classmethod
    def save_to_feature_config(self, workspace_path: Path, config_save_dir: Path):
        """Save feature definition within the workspace into HOCON feature config files"""
        repo_definitions = self._extract_features(workspace_path)
        self._save_request_feature_config(repo_definitions, config_save_dir)
        self._save_anchored_feature_config(repo_definitions, config_save_dir)
        self._save_derived_feature_config(repo_definitions, config_save_dir)

    @classmethod
    def save_to_feature_config_from_context(self, anchor_list, derived_feature_list, local_workspace_dir: Path):
        """Save feature definition within the workspace into HOCON feature config files from current context, rather than reading from python files"""
        repo_definitions = self._extract_features_from_context(
            anchor_list, derived_feature_list, local_workspace_dir)
        self._save_request_feature_config(
            repo_definitions, local_workspace_dir)
        self._save_anchored_feature_config(
            repo_definitions, local_workspace_dir)
        self._save_derived_feature_config(
            repo_definitions, local_workspace_dir)

    @classmethod
    def _save_request_feature_config(self, repo_definitions: RepoDefinitions, local_workspace_dir="./"):
        config_file_name = "feature_conf/auto_generated_request_features.conf"
        tm = Template(
            """
// THIS FILE IS AUTO GENERATED. PLEASE DO NOT EDIT.
anchors: {
    {% for anchor in feature_anchors %}
        {% if anchor.source.name == "PASSTHROUGH" %}
            {{anchor.to_feature_config()}}
        {% endif %}
    {% endfor %}
}
"""
        )

        request_feature_configs = tm.render(
            feature_anchors=repo_definitions.feature_anchors)
        config_file_path = os.path.join(local_workspace_dir, config_file_name)
        write_to_file(content=request_feature_configs,
                      full_file_name=config_file_path)

    @classmethod
    def _save_anchored_feature_config(self, repo_definitions: RepoDefinitions, local_workspace_dir="./"):
        config_file_name = "feature_conf/auto_generated_anchored_features.conf"
        tm = Template(
            """
// THIS FILE IS AUTO GENERATED. PLEASE DO NOT EDIT.
anchors: {
    {% for anchor in feature_anchors %}
        {% if not anchor.source.name == "PASSTHROUGH" %}
            {{anchor.to_feature_config()}}
        {% endif %}
    {% endfor %}
}

sources: {
    {% for source in sources%}
        {% if not source.name == "PASSTHROUGH" %}
            {{source.to_feature_config()}}
        {% endif %}
    {% endfor %}
}
"""
        )
        anchored_feature_configs = tm.render(feature_anchors=repo_definitions.feature_anchors,
                                             sources=repo_definitions.sources)
        config_file_path = os.path.join(local_workspace_dir, config_file_name)
        write_to_file(content=anchored_feature_configs,
                      full_file_name=config_file_path)

    @classmethod
    def _save_derived_feature_config(self, repo_definitions: RepoDefinitions, local_workspace_dir="./"):
        config_file_name = "feature_conf/auto_generated_derived_features.conf"
        tm = Template(
            """
anchors: {}
derivations: {
    {% for derived_feature in derived_features %}
        {{derived_feature.to_feature_config()}}
    {% endfor %}
}
"""
        )
        derived_feature_configs = tm.render(
            derived_features=repo_definitions.derived_features)
        config_file_path = os.path.join(local_workspace_dir, config_file_name)
        write_to_file(content=derived_feature_configs,
                      full_file_name=config_file_path)


def check(r):
    if not r.ok:
        raise RuntimeError(
            f"Failed to call registry API, status is {r.status_code}, error is {r.text}")
    return r


def source_to_def(v: Source) -> dict:
    ret = {}
    if v.name == INPUT_CONTEXT:
        return {
            "name": INPUT_CONTEXT,
            "type": INPUT_CONTEXT,
            "path": INPUT_CONTEXT,
        }
    elif isinstance(v, HdfsSource):
        ret = {
            "name": v.name,
            "type": urlparse(v.path).scheme,
            "path": v.path,
        }
    elif isinstance(v, JdbcSource):
        ret = {
            "name": v.name,
            "type": "jdbc",
            "url": v.url,
        }
        if hasattr(v, "dbtable") and v.dbtable:
            ret["dbtable"] = v.dbtable
        if hasattr(v, "query") and v.query:
            ret["query"] = v.query
        if hasattr(v, "auth") and v.auth:
            ret["auth"] = v.auth
    elif isinstance(v, GenericSource):
        ret = v.to_dict()
        ret["name"] = v.name
    else:
        raise ValueError(f"Unsupported source type {v.__class__}")
    if hasattr(v, "preprocessing") and v.preprocessing:
        ret["preprocessing"] = inspect.getsource(v.preprocessing)
    if v.event_timestamp_column:
        ret["eventTimestampColumn"] = v.event_timestamp_column
        ret["event_timestamp_column"] = v.event_timestamp_column
    if v.timestamp_format:
        ret["timestampFormat"] = v.timestamp_format
        ret["timestamp_format"] = v.timestamp_format
    if v.registry_tags:
        ret["tags"] = v.registry_tags
    return ret


def dict_to_source(v: dict) -> Source:
    id = UUID(v["guid"])
    type = v["attributes"]["type"]
    source = None
    if type == INPUT_CONTEXT:
        source = InputContext()
    elif type == "jdbc":
        source = JdbcSource(name=v["attributes"]["name"],
                            url=v["attributes"].get("url"),
                            dbtable=v["attributes"].get("dbtable"),
                            query=v["attributes"].get("query"),
                            auth=v["attributes"].get("auth"),
                            preprocessing=_correct_function_indentation(
                                v["attributes"].get("preprocessing")),
                            event_timestamp_column=v["attributes"].get(
                                "eventTimestampColumn"),
                            timestamp_format=v["attributes"].get(
                                "timestampFormat"),
                            registry_tags=v["attributes"].get("tags", {}))
    elif type == "generic":
        options = v["attributes"].copy()
        # These are not options
        del options["type"]
        del options["name"]
        del options["format"]
        if "qualifiedName" in options:
            del options["qualifiedName"]
        if "mode" in options:
            del options["mode"]
        if "preprocessing" in options:
            del options["preprocessing"]
        if "eventTimestampColumn" in options:
            del options["eventTimestampColumn"]
        if "timestampFormat" in options:
            del options["timestampFormat"]
        if "tags" in options:
            del options["tags"]
        source = GenericSource(
            name=v["attributes"]["name"],
            format=v["attributes"]["format"],
            mode=v["attributes"].get("mode", ""),
            options=options,
            preprocessing=_correct_function_indentation(
                v["attributes"].get("preprocessing")),
            event_timestamp_column=v["attributes"].get(
                "eventTimestampColumn"),
            timestamp_format=v["attributes"].get(
                "timestampFormat"),
            registry_tags=v["attributes"].get("tags", {}))
    elif v["attributes"].get("path"):
        source = HdfsSource(name=v["attributes"]["name"],
                            path=v["attributes"]["path"],
                            preprocessing=_correct_function_indentation(
                                v["attributes"].get("preprocessing")),
                            event_timestamp_column=v["attributes"].get(
                                "eventTimestampColumn"),
                            timestamp_format=v["attributes"].get(
                                "timestampFormat"),
                            registry_tags=v["attributes"].get("tags", {}))
    else:
        raise ValueError(f"Invalid source format {type}")
    source._registry_id = id
    source._qualified_name = v["attributes"]["qualifiedName"]
    return source


def anchor_to_def(v: FeatureAnchor) -> dict:
    source_id = v.source._registry_id
    ret = {
        "name": v.name,
        "sourceId": str(source_id),
    }
    if v.registry_tags:
        ret["tags"] = v.registry_tags
    return ret


def dict_to_anchor(v: dict) -> FeatureAnchor:
    ret = FeatureAnchor(name=v["attributes"]["name"],
                        source=None,
                        features=[],
                        registry_tags=v["attributes"].get("tags", {}),
                        __no_validate=True)
    ret._source_id = UUID(v["attributes"]["source"]["guid"])
    ret._features = [UUID(f["guid"]) for f in v["attributes"]["features"]]
    ret._qualified_name = v["attributes"]["qualifiedName"]
    ret._registry_id = UUID(v["guid"])
    return ret


def transformation_to_def(v: Transformation) -> dict:
    if isinstance(v, ExpressionTransformation):
        return {
            "transformExpr": v.expr
        }
    elif isinstance(v, WindowAggTransformation):
        ret = {
            "defExpr": v.def_expr,
        }
        if v.agg_func:
            ret["aggFunc"] = v.agg_func
        if v.window:
            ret["window"] = v.window
        if v.group_by:
            ret["groupBy"] = v.group_by
        if v.filter:
            ret["filter"] = v.filter
        if v.limit:
            ret["limit"] = v.limit
        return ret
    raise ValueError("Unsupported Transformation type")


def dict_to_transformation(v: dict) -> Transformation:
    if v is None:
        return None
    v = to_camel(v)
    if 'transformExpr' in v:
        # it's ExpressionTransformation
        return ExpressionTransformation(v['transformExpr'])
    elif 'defExpr' in v:
        return WindowAggTransformation(agg_expr=v['defExpr'],
                                       agg_func=v.get('aggFunc'),
                                       window=v.get('window'),
                                       group_by=v.get('groupBy'),
                                       filter=v.get('filter'),
                                       limit=v.get('limit'))
    raise ValueError(f"Invalid transformation format {v}")


def feature_type_to_def(v: FeatureType) -> dict:
    return {
        "type": v.type,
        "tensorCategory": v.tensor_category,
        "dimensionType": [value_type_to_str(t) for t in v.dimension_type],
        "valType": value_type_to_str(v.val_type),
    }


def dict_to_feature_type(v: dict) -> FeatureType:
    return FeatureType(val_type=str_to_value_type(v["valType"]),
                       dimension_type=[str_to_value_type(
                           s) for s in v["dimensionType"]],
                       tensor_category=v["tensorCategory"],
                       type=v["type"])


def typed_key_to_def(v: TypedKey) -> dict:
    ret = {
        "keyColumn": v.key_column,
        "keyColumnType": value_type_to_str(v.key_column_type)
    }
    if v.full_name:
        ret["fullName"] = v.full_name
    if v.description:
        ret["description"] = v.description
    if v.key_column_alias:
        ret["keyColumnAlias"] = v.key_column_alias
    return ret


def dict_to_typed_key(v: dict) -> TypedKey:
    v = to_camel(v)
    return TypedKey(key_column=v["keyColumn"],
                    key_column_type=str_to_value_type(v["keyColumnType"]),
                    full_name=v.get("fullName"),
                    description=v.get("description"),
                    key_column_alias=v.get("keyColumnAlias"))


def feature_to_def(v: Feature) -> dict:
    ret = {
        "name": v.name,
        "featureType": feature_type_to_def(v.feature_type),
        "key": [typed_key_to_def(k) for k in v.key],
    }
    if v.transform:
        ret["transformation"] = transformation_to_def(
            v.transform)
    if v.registry_tags:
        ret["tags"] = v.registry_tags
    return ret


def dict_to_feature(v: dict) -> Feature:
    ret = Feature(name=v["attributes"]["name"],
                  feature_type=dict_to_feature_type(v["attributes"]["type"]),
                  key=[dict_to_typed_key(k) for k in v["attributes"]["key"]],
                  transform=dict_to_transformation(
                      v["attributes"].get("transformation")),
                  registry_tags=v["attributes"].get("tags", {}))
    ret._qualified_name = v["attributes"]["qualifiedName"]
    ret._registry_id = UUID(v["guid"])
    return ret


def _get_type_name(v: Any) -> str:
    if isinstance(v, Source):
        return TYPEDEF_SOURCE
    elif isinstance(v, FeatureAnchor):
        return TYPEDEF_ANCHOR
    elif isinstance(v, DerivedFeature):
        return TYPEDEF_DERIVED_FEATURE
    elif isinstance(v, Feature):
        return TYPEDEF_ANCHOR_FEATURE
    raise TypeError("Invalid type")


def derived_feature_to_def(v: DerivedFeature) -> dict:
    ret = {
        "name": v.name,
        "featureType": feature_type_to_def(v.feature_type),
        "key": [typed_key_to_def(k) for k in v.key],
        "inputAnchorFeatures": [str(f._registry_id) for f in v.input_features if not isinstance(f, DerivedFeature)],
        "inputDerivedFeatures": [str(f._registry_id) for f in v.input_features if isinstance(f, DerivedFeature)],
    }
    if v.transform:
        ret["transformation"] = transformation_to_def(v.transform)
    return ret


def dict_to_derived_feature(v: dict) -> DerivedFeature:
    v["attributes"] = to_camel(v["attributes"])
    ret = DerivedFeature(name=v["attributes"]["name"],
                         feature_type=dict_to_feature_type(
                             v["attributes"]["type"]),
                         input_features=[],
                         key=[dict_to_typed_key(k)
                              for k in v["attributes"]["key"]],
                         transform=dict_to_transformation(
                             v["attributes"]["transformation"]),
                         registry_tags=v["attributes"].get("tags", {}),
                         __no_validate=True)
    ret._input_anchor_features = [
        UUID(f["guid"]) for f in v["attributes"]["inputAnchorFeatures"]]
    ret._input_derived_features = [
        UUID(f["guid"]) for f in v["attributes"]["inputDerivedFeatures"]]
    ret._qualified_name = v["attributes"]["qualifiedName"]
    ret._registry_id = UUID(v["guid"])
    return ret


class Relation:
    def __init__(self, v: dict):
        self.f = UUID(v["fromEntityId"])
        self.t = UUID(v["toEntityId"])
        self.type = v["relationshipType"]


def dict_to_project(v: dict) -> Tuple[List[FeatureAnchor], List[DerivedFeature]]:
    entities = v["guidEntityMap"]
    # Step 1, Extract each entity
    sources = dict([(UUID(k), dict_to_source(entities[k]))
                   for k in entities if entities[k]["typeName"] == "feathr_source_v1"])
    anchors = dict([(UUID(k), dict_to_anchor(entities[k]))
                   for k in entities if entities[k]["typeName"] == "feathr_anchor_v1"])
    features = dict([(UUID(k), dict_to_feature(entities[k]))
                    for k in entities if entities[k]["typeName"] == "feathr_anchor_feature_v1"])
    derived_features = dict([(UUID(k), dict_to_derived_feature(entities[k]))
                            for k in entities if entities[k]["typeName"] == "feathr_derived_feature_v1"])
    # Step 2, Setup connections between extracted entities
    # Step 2-1, Set up anchors
    for k in anchors:
        anchor = anchors[k]
        anchor.source = sources[anchor._source_id]
        anchor.features = [features[id] for id in anchor._features]
    # Step 2-1, Set up derived features
    for k in derived_features:
        df = derived_features[k]
        input_anchor_features = [features[id]
                                 for id in df._input_anchor_features]
        input_derived_features = [derived_features[id]
                                  for id in df._input_derived_features]
        df.input_features = input_anchor_features + input_derived_features
    return (list(anchors.values()), list(derived_features.values()))


def _correct_function_indentation(user_func: str) -> str:
    """
    The function read from registry might have the wrong indentation. We need to correct those indentations.
    More specifically, we are using the inspect module to copy the function body for UDF for further submission. In that case, there will be situations like this:

    def feathr_udf1(df)
        return df

            def feathr_udf2(df)
                return df

    For example, in Feathr test cases, there are similar patterns for `feathr_udf2`, since it's defined in another function body (the registry_test_setup method).
    This is not an ideal way of dealing with that, but we'll keep it here until we figure out a better way.
    """
    if user_func is None:
        return None
    # if user_func is a string, turn it into a list of strings so that it can be used below
    temp_udf_source_code = user_func.split('\n')
    # assuming the first line is the function name
    leading_space_num = len(
        temp_udf_source_code[0]) - len(temp_udf_source_code[0].lstrip())
    # strip the lines to make sure the function has the correct indentation
    udf_source_code_striped = [line[leading_space_num:]
                               for line in temp_udf_source_code]
    # append '\n' back since it was deleted due to the previous split
    udf_source_code = [line+'\n' for line in udf_source_code_striped]
    return " ".join(udf_source_code)



from typing import Optional
from jinja2 import Template
from loguru import logger
from feathr.definition.feathrconfig import HoconConvertible

class ObservationSettings(HoconConvertible):
    """Time settings of the observation data. Used in feature join.

    Attributes:
        observation_path: path to the observation dataset, i.e. input dataset to get with features
        event_timestamp_column (Optional[str]): The timestamp field of your record. As sliding window aggregation feature assume each record in the source data should have a timestamp column.
        timestamp_format (Optional[str], optional): The format of the timestamp field. Defaults to "epoch". Possible values are:
            - `epoch` (seconds since epoch), for example `1647737463`
            - `epoch_millis` (milliseconds since epoch), for example `1647737517761`
            - Any date formats supported by [SimpleDateFormat](https://docs.oracle.com/javase/8/docs/api/java/text/SimpleDateFormat.html).
    """
    def __init__(self,
                 observation_path: str,
                 event_timestamp_column: Optional[str] = None,
                 timestamp_format: str = "epoch") -> None:
        self.event_timestamp_column = event_timestamp_column
        self.timestamp_format = timestamp_format
        self.observation_path = observation_path
        if observation_path.startswith("http"):
            logger.warning("Your observation_path {} starts with http, which is not supported. Consider using paths starting with wasb[s]/abfs[s]/s3.", observation_path)

    def to_feature_config(self) -> str:
        tm = Template("""
                {% if setting.event_timestamp_column is not none %}
                settings: {
                    joinTimeSettings: {
                        timestampColumn: {
                            def: "{{setting.event_timestamp_column}}"
                            format: "{{setting.timestamp_format}}"
                        }
                    }
                }
                {% endif %}
                observationPath: "{{setting.observation_path}}"
            """)
        return tm.render(setting=self)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: featureValue.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x12\x66\x65\x61tureValue.proto\x12\x08protobuf\"\xce\x06\n\x0c\x46\x65\x61tureValue\x12\x17\n\rboolean_value\x18\x01 \x01(\x08H\x00\x12\x16\n\x0cstring_value\x18\x02 \x01(\tH\x00\x12\x15\n\x0b\x66loat_value\x18\x03 \x01(\x02H\x00\x12\x16\n\x0c\x64ouble_value\x18\x04 \x01(\x01H\x00\x12\x13\n\tint_value\x18\x05 \x01(\x05H\x00\x12\x14\n\nlong_value\x18\x06 \x01(\x03H\x00\x12/\n\rboolean_array\x18\n \x01(\x0b\x32\x16.protobuf.BooleanArrayH\x00\x12-\n\x0cstring_array\x18\x0b \x01(\x0b\x32\x15.protobuf.StringArrayH\x00\x12+\n\x0b\x66loat_array\x18\x0c \x01(\x0b\x32\x14.protobuf.FloatArrayH\x00\x12-\n\x0c\x64ouble_array\x18\r \x01(\x0b\x32\x15.protobuf.DoubleArrayH\x00\x12+\n\tint_array\x18\x0e \x01(\x0b\x32\x16.protobuf.IntegerArrayH\x00\x12)\n\nlong_array\x18\x0f \x01(\x0b\x32\x13.protobuf.LongArrayH\x00\x12*\n\nbyte_array\x18\x10 \x01(\x0b\x32\x14.protobuf.BytesArrayH\x00\x12:\n\x13sparse_string_array\x18\x14 \x01(\x0b\x32\x1b.protobuf.SparseStringArrayH\x00\x12\x36\n\x11sparse_bool_array\x18\x15 \x01(\x0b\x32\x19.protobuf.SparseBoolArrayH\x00\x12<\n\x14sparse_integer_array\x18\x16 \x01(\x0b\x32\x1c.protobuf.SparseIntegerArrayH\x00\x12\x36\n\x11sparse_long_array\x18\x17 \x01(\x0b\x32\x19.protobuf.SparseLongArrayH\x00\x12:\n\x13sparse_double_array\x18\x18 \x01(\x0b\x32\x1b.protobuf.SparseDoubleArrayH\x00\x12\x38\n\x12sparse_float_array\x18\x19 \x01(\x0b\x32\x1a.protobuf.SparseFloatArrayH\x00\x42\x13\n\x11\x46\x65\x61tureValueOneOf\" \n\x0c\x42ooleanArray\x12\x10\n\x08\x62ooleans\x18\x01 \x03(\x08\"\x1e\n\x0bStringArray\x12\x0f\n\x07strings\x18\x01 \x03(\t\"\x1e\n\x0b\x44oubleArray\x12\x0f\n\x07\x64oubles\x18\x01 \x03(\x01\"\x1c\n\nFloatArray\x12\x0e\n\x06\x66loats\x18\x01 \x03(\x02\" \n\x0cIntegerArray\x12\x10\n\x08integers\x18\x01 \x03(\x05\"\x1a\n\tLongArray\x12\r\n\x05longs\x18\x01 \x03(\x03\"\x1b\n\nBytesArray\x12\r\n\x05\x62ytes\x18\x01 \x03(\x0c\"B\n\x11SparseStringArray\x12\x16\n\x0eindex_integers\x18\x01 \x03(\x05\x12\x15\n\rvalue_strings\x18\x02 \x03(\t\"A\n\x0fSparseBoolArray\x12\x16\n\x0eindex_integers\x18\x01 \x03(\x05\x12\x16\n\x0evalue_booleans\x18\x02 \x03(\x08\"D\n\x12SparseIntegerArray\x12\x16\n\x0eindex_integers\x18\x01 \x03(\x05\x12\x16\n\x0evalue_integers\x18\x02 \x03(\x05\">\n\x0fSparseLongArray\x12\x16\n\x0eindex_integers\x18\x01 \x03(\x05\x12\x13\n\x0bvalue_longs\x18\x02 \x03(\x03\"B\n\x11SparseDoubleArray\x12\x16\n\x0eindex_integers\x18\x01 \x03(\x05\x12\x15\n\rvalue_doubles\x18\x02 \x03(\x01\"@\n\x10SparseFloatArray\x12\x16\n\x0eindex_integers\x18\x01 \x03(\x05\x12\x14\n\x0cvalue_floats\x18\x02 \x03(\x02\x42+\n)com.linkedin.feathr.common.types.protobufb\x06proto3')



_FEATUREVALUE = DESCRIPTOR.message_types_by_name['FeatureValue']
_BOOLEANARRAY = DESCRIPTOR.message_types_by_name['BooleanArray']
_STRINGARRAY = DESCRIPTOR.message_types_by_name['StringArray']
_DOUBLEARRAY = DESCRIPTOR.message_types_by_name['DoubleArray']
_FLOATARRAY = DESCRIPTOR.message_types_by_name['FloatArray']
_INTEGERARRAY = DESCRIPTOR.message_types_by_name['IntegerArray']
_LONGARRAY = DESCRIPTOR.message_types_by_name['LongArray']
_BYTESARRAY = DESCRIPTOR.message_types_by_name['BytesArray']
_SPARSESTRINGARRAY = DESCRIPTOR.message_types_by_name['SparseStringArray']
_SPARSEBOOLARRAY = DESCRIPTOR.message_types_by_name['SparseBoolArray']
_SPARSEINTEGERARRAY = DESCRIPTOR.message_types_by_name['SparseIntegerArray']
_SPARSELONGARRAY = DESCRIPTOR.message_types_by_name['SparseLongArray']
_SPARSEDOUBLEARRAY = DESCRIPTOR.message_types_by_name['SparseDoubleArray']
_SPARSEFLOATARRAY = DESCRIPTOR.message_types_by_name['SparseFloatArray']
FeatureValue = _reflection.GeneratedProtocolMessageType('FeatureValue', (_message.Message,), {
  'DESCRIPTOR' : _FEATUREVALUE,
  '__module__' : 'featureValue_pb2'
  # @@protoc_insertion_point(class_scope:protobuf.FeatureValue)
})
_sym_db.RegisterMessage(FeatureValue)

BooleanArray = _reflection.GeneratedProtocolMessageType('BooleanArray', (_message.Message,), {
  'DESCRIPTOR' : _BOOLEANARRAY,
  '__module__' : 'featureValue_pb2'
  # @@protoc_insertion_point(class_scope:protobuf.BooleanArray)
})
_sym_db.RegisterMessage(BooleanArray)

StringArray = _reflection.GeneratedProtocolMessageType('StringArray', (_message.Message,), {
  'DESCRIPTOR' : _STRINGARRAY,
  '__module__' : 'featureValue_pb2'
  # @@protoc_insertion_point(class_scope:protobuf.StringArray)
})
_sym_db.RegisterMessage(StringArray)

DoubleArray = _reflection.GeneratedProtocolMessageType('DoubleArray', (_message.Message,), {
  'DESCRIPTOR' : _DOUBLEARRAY,
  '__module__' : 'featureValue_pb2'
  # @@protoc_insertion_point(class_scope:protobuf.DoubleArray)
})
_sym_db.RegisterMessage(DoubleArray)

FloatArray = _reflection.GeneratedProtocolMessageType('FloatArray', (_message.Message,), {
  'DESCRIPTOR' : _FLOATARRAY,
  '__module__' : 'featureValue_pb2'
  # @@protoc_insertion_point(class_scope:protobuf.FloatArray)
})
_sym_db.RegisterMessage(FloatArray)

IntegerArray = _reflection.GeneratedProtocolMessageType('IntegerArray', (_message.Message,), {
  'DESCRIPTOR' : _INTEGERARRAY,
  '__module__' : 'featureValue_pb2'
  # @@protoc_insertion_point(class_scope:protobuf.IntegerArray)
})
_sym_db.RegisterMessage(IntegerArray)

LongArray = _reflection.GeneratedProtocolMessageType('LongArray', (_message.Message,), {
  'DESCRIPTOR' : _LONGARRAY,
  '__module__' : 'featureValue_pb2'
  # @@protoc_insertion_point(class_scope:protobuf.LongArray)
})
_sym_db.RegisterMessage(LongArray)

BytesArray = _reflection.GeneratedProtocolMessageType('BytesArray', (_message.Message,), {
  'DESCRIPTOR' : _BYTESARRAY,
  '__module__' : 'featureValue_pb2'
  # @@protoc_insertion_point(class_scope:protobuf.BytesArray)
})
_sym_db.RegisterMessage(BytesArray)

SparseStringArray = _reflection.GeneratedProtocolMessageType('SparseStringArray', (_message.Message,), {
  'DESCRIPTOR' : _SPARSESTRINGARRAY,
  '__module__' : 'featureValue_pb2'
  # @@protoc_insertion_point(class_scope:protobuf.SparseStringArray)
})
_sym_db.RegisterMessage(SparseStringArray)

SparseBoolArray = _reflection.GeneratedProtocolMessageType('SparseBoolArray', (_message.Message,), {
  'DESCRIPTOR' : _SPARSEBOOLARRAY,
  '__module__' : 'featureValue_pb2'
  # @@protoc_insertion_point(class_scope:protobuf.SparseBoolArray)
})
_sym_db.RegisterMessage(SparseBoolArray)

SparseIntegerArray = _reflection.GeneratedProtocolMessageType('SparseIntegerArray', (_message.Message,), {
  'DESCRIPTOR' : _SPARSEINTEGERARRAY,
  '__module__' : 'featureValue_pb2'
  # @@protoc_insertion_point(class_scope:protobuf.SparseIntegerArray)
})
_sym_db.RegisterMessage(SparseIntegerArray)

SparseLongArray = _reflection.GeneratedProtocolMessageType('SparseLongArray', (_message.Message,), {
  'DESCRIPTOR' : _SPARSELONGARRAY,
  '__module__' : 'featureValue_pb2'
  # @@protoc_insertion_point(class_scope:protobuf.SparseLongArray)
})
_sym_db.RegisterMessage(SparseLongArray)

SparseDoubleArray = _reflection.GeneratedProtocolMessageType('SparseDoubleArray', (_message.Message,), {
  'DESCRIPTOR' : _SPARSEDOUBLEARRAY,
  '__module__' : 'featureValue_pb2'
  # @@protoc_insertion_point(class_scope:protobuf.SparseDoubleArray)
})
_sym_db.RegisterMessage(SparseDoubleArray)

SparseFloatArray = _reflection.GeneratedProtocolMessageType('SparseFloatArray', (_message.Message,), {
  'DESCRIPTOR' : _SPARSEFLOATARRAY,
  '__module__' : 'featureValue_pb2'
  # @@protoc_insertion_point(class_scope:protobuf.SparseFloatArray)
})
_sym_db.RegisterMessage(SparseFloatArray)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n)com.linkedin.feathr.common.types.protobuf'
  _FEATUREVALUE._serialized_start=33
  _FEATUREVALUE._serialized_end=879
  _BOOLEANARRAY._serialized_start=881
  _BOOLEANARRAY._serialized_end=913
  _STRINGARRAY._serialized_start=915
  _STRINGARRAY._serialized_end=945
  _DOUBLEARRAY._serialized_start=947
  _DOUBLEARRAY._serialized_end=977
  _FLOATARRAY._serialized_start=979
  _FLOATARRAY._serialized_end=1007
  _INTEGERARRAY._serialized_start=1009
  _INTEGERARRAY._serialized_end=1041
  _LONGARRAY._serialized_start=1043
  _LONGARRAY._serialized_end=1069
  _BYTESARRAY._serialized_start=1071
  _BYTESARRAY._serialized_end=1098
  _SPARSESTRINGARRAY._serialized_start=1100
  _SPARSESTRINGARRAY._serialized_end=1166
  _SPARSEBOOLARRAY._serialized_start=1168
  _SPARSEBOOLARRAY._serialized_end=1233
  _SPARSEINTEGERARRAY._serialized_start=1235
  _SPARSEINTEGERARRAY._serialized_end=1303
  _SPARSELONGARRAY._serialized_start=1305
  _SPARSELONGARRAY._serialized_end=1367
  _SPARSEDOUBLEARRAY._serialized_start=1369
  _SPARSEDOUBLEARRAY._serialized_end=1435
  _SPARSEFLOATARRAY._serialized_start=1437
  _SPARSEFLOATARRAY._serialized_end=1501
# @@protoc_insertion_point(module_scope)



from typing import List, Optional, Dict
from feathr.definition.feature import Feature
from feathr.definition.source import Source
from feathr.definition.typed_key import DUMMY_KEY
from jinja2 import Template
from feathr.definition.source import INPUT_CONTEXT
from feathr.definition.feathrconfig import HoconConvertible


class FeatureAnchor(HoconConvertible):
    """
        A feature anchor defines a set of features on top of a data source, a.k.a. a set of features anchored to a source.

        The feature producer writes multiple anchors for a feature, exposing the same feature name for the feature
        consumer to reference it.

        Attributes:
            name: Unique name of the anchor.
            source: data source that the features are anchored to. Should be either of `INPUT_CONTEXT` or `feathr.source.Source`
            features: list of features defined within this anchor.
            registry_tags: A dict of (str, str) that you can pass to feature registry for better organization.
                           For example, you can use {"deprecated": "true"} to indicate this anchor is deprecated, etc.
    """
    def __init__(self,
                name: str,
                source: Source,
                features: List[Feature],
                registry_tags: Optional[Dict[str, str]] = None,
                **kwargs):
        self.name = name
        self.features = features
        self.source = source
        self.registry_tags=registry_tags
        # Add a hidden option to skip validation, Anchor could be half-constructed during the loading from registry
        if not kwargs.get("__no_validate", False) :
            self.validate_features()

    def validate_features(self):
        """Validate that anchor is non-empty and all its features share the same key"""
        # assert len(self.features) > 0
        if self.source != INPUT_CONTEXT:
            for feature in self.features:
                if feature.key == [DUMMY_KEY]:
                    raise RuntimeError(f"For anchors of non-INPUT_CONTEXT source, key of feature {feature.name} "
                                       f"should be explicitly specified and not left blank.")
        
    def to_feature_config(self) -> str:
        tm = Template("""
            {{anchor_name}}: {
                source: {{source.name}}
                key.sqlExpr: [{{key_list}}]
                features: {
                    {% for feature in features %}
                        {{feature.to_feature_config()}}
                    {% endfor %}
                }
            }
        """)
        key_list = ','.join((key for key in self.features[0].key_alias) if len(self.features)!=0 else [])
        return tm.render(anchor_name = self.name,
                        key_list = key_list,
                        features = self.features,
                        source = self.source)

    def __str__(self):
        return self.to_feature_config()



from abc import ABC, abstractmethod
from pathlib import Path

from typing import Any, Dict, List, Optional, Tuple
from feathr.definition.feature_derivations import DerivedFeature
from feathr.definition.anchor import FeatureAnchor
from feathr.utils._envvariableutil import _EnvVaraibleUtil

class FeathrRegistry(ABC):
    """This is the abstract class for all the feature registries. All the feature registries should implement those interfaces.
    """

    @abstractmethod
    def register_features(self, anchor_list: List[FeatureAnchor] =[], derived_feature_list: List[DerivedFeature]=[]):
        """Registers features based on the current workspace

                Args:
                anchor_list: List of FeatureAnchors
                derived_feature_list: List of DerivedFeatures
        """
        pass


    @abstractmethod
    def list_registered_features(self, project_name: str) -> List[str]:
        """List all the already registered features under the given project.
        `project_name` must not be None or empty string because it violates the RBAC policy
        """
        pass

    @abstractmethod
    def get_features_from_registry(self, project_name: str) -> Tuple[List[FeatureAnchor], List[DerivedFeature]]:
        """[Sync Features from registry to local workspace, given a project_name, will write project's features from registry to to user's local workspace]

        Args:
            project_name (str): project name.

        Returns:
            bool: Returns true if the job completed successfully, otherwise False
        """
        pass
    
    @classmethod
    @abstractmethod
    def save_to_feature_config(self, workspace_path: Path, config_save_dir: Path):
        """Save feature definition within the workspace into HOCON feature config files"""
        pass

    @classmethod
    @abstractmethod
    def save_to_feature_config_from_context(self, anchor_list, derived_feature_list, local_workspace_dir: Path):
        """Save feature definition within the workspace into HOCON feature config files from current context, rather than reading from python files"""
        pass

def default_registry_client(project_name: str, config_path:str = "./feathr_config.yaml", project_registry_tag: Dict[str, str]=None, credential = None) -> FeathrRegistry:
    from feathr.registry._feathr_registry_client import _FeatureRegistry
    from feathr.registry._feature_registry_purview import _PurviewRegistry
    envutils = _EnvVaraibleUtil(config_path)
    registry_endpoint = envutils.get_environment_variable_with_default("feature_registry", "api_endpoint")
    if registry_endpoint:
        return _FeatureRegistry(project_name, endpoint=registry_endpoint, project_tags=project_registry_tag, credential=credential)
    else:
        registry_delimiter = envutils.get_environment_variable_with_default('feature_registry', 'purview', 'delimiter')
        azure_purview_name = envutils.get_environment_variable_with_default('feature_registry', 'purview', 'purview_name')
        # initialize the registry no matter whether we set purview name or not, given some of the methods are used there.
        return _PurviewRegistry(project_name, azure_purview_name, registry_delimiter, project_registry_tag, config_path = config_path, credential=credential)



import os
from datetime import datetime, timedelta
from pathlib import Path

from feathr import (BackfillTime, MaterializationSettings)
# from feathr import *
from feathr.client import FeathrClient
from feathr.definition.dtype import ValueType
from feathr.definition.query_feature_list import FeatureQuery
from feathr.definition.settings import ObservationSettings
from feathr.definition.typed_key import TypedKey
from test_fixture import (basic_test_setup, get_online_test_table_name)
from test_utils.constants import Constants

def test_feathr_online_store_agg_features():
    """
    Test FeathrClient() get_online_features and batch_get can get data correctly.
    """

    online_test_table = get_online_test_table_name("nycTaxiCITableMaven")
    test_workspace_dir = Path(
        __file__).parent.resolve() / "test_user_workspace"
    # os.chdir(test_workspace_dir)

    # The `feathr_runtime_location` was commented out in this config file, so feathr should use
    # Maven package as the dependency and `noop.jar` as the main file
    client: FeathrClient = basic_test_setup(os.path.join(test_workspace_dir, "feathr_config_maven.yaml"))

    
    
    location_id = TypedKey(key_column="DOLocationID",
                            key_column_type=ValueType.INT32,
                            description="location id in NYC",
                            full_name="nyc_taxi.location_id")

    feature_query = FeatureQuery(
        feature_list=["f_location_avg_fare"], key=location_id)
    settings = ObservationSettings(
        observation_path="wasbs://public@azurefeathrstorage.blob.core.windows.net/sample_data/green_tripdata_2020-04.csv",
        event_timestamp_column="lpep_dropoff_datetime",
        timestamp_format="yyyy-MM-dd HH:mm:ss")

    now = datetime.now()
    # set output folder based on different runtime
    if client.spark_runtime == 'databricks':
        output_path = ''.join(['dbfs:/feathrazure_cijob','_', str(now.minute), '_', str(now.second), ".avro"])
    else:
        output_path = ''.join(['abfss://feathrazuretest3fs@feathrazuretest3storage.dfs.core.windows.net/demo_data/output','_', str(now.minute), '_', str(now.second), ".avro"])


    client.get_offline_features(observation_settings=settings,
                                feature_query=feature_query,
                                output_path=output_path)

    # assuming the job can successfully run; otherwise it will throw exception
    client.wait_job_to_finish(timeout_sec=Constants.SPARK_JOB_TIMEOUT_SECONDS)
    return
    backfill_time = BackfillTime(start=datetime(
        2020, 5, 20), end=datetime(2020, 5, 20), step=timedelta(days=1))
    redisSink = RedisSink(table_name=online_test_table)
    settings = MaterializationSettings("TestJobName",
                                       sinks=[redisSink],
                                       feature_names=[
                                           "f_location_avg_fare", "f_location_max_fare"],
                                       backfill_time=backfill_time)
    client.materialize_features(settings)
    # just assume the job is successful without validating the actual result in Redis. Might need to consolidate
    # this part with the test_feathr_online_store test case
    client.wait_job_to_finish(timeout_sec=Constants.SPARK_JOB_TIMEOUT_SECONDS)

    res = client.get_online_features(online_test_table, '265', [
                                     'f_location_avg_fare', 'f_location_max_fare'])
    # just assume there are values. We don't hard code the values for now for testing
    # the correctness of the feature generation should be guaranteed by feathr runtime.
    # ID 239 and 265 are available in the `DOLocationID` column in this file:
    # https://s3.amazonaws.com/nyc-tlc/trip+data/green_tripdata_2020-04.csv
    # View more details on this dataset: https://www1.nyc.gov/site/tlc/about/tlc-trip-record-data.page
    assert len(res) == 2
    assert res[0] != None
    assert res[1] != None
    res = client.multi_get_online_features(online_test_table,
                                           ['239', '265'],
                                           ['f_location_avg_fare', 'f_location_max_fare'])
    assert res['239'][0] != None
    assert res['239'][1] != None
    assert res['265'][0] != None
    assert res['265'][1] != None



from __future__ import annotations
from copy import copy, deepcopy

from typing import List, Optional
from feathr.definition.dtype import ValueType, FeatureType


class TypedKey:
    """The key of a feature. A feature is typically keyed by some id(s). e.g. product id, user id
      Attributes:
        key_column: The id column name of this key. e.g. 'product_id'.
        key_column_type: Types of the key_column
        full_name: Unique name of the key. Recommend using [project_name].[key_name], e.g. ads.user_id
        description: Documentation for the key.
        key_column_alias: Used in some advanced derived features. Default to the key_column.
    """
    def __init__(self,
                 key_column: str,
                 key_column_type: ValueType,
                 full_name: Optional[str] = None,
                 description: Optional[str] = None,
                 key_column_alias: Optional[str] = None) -> None:
        self.key_column = key_column
        self.key_column_type = key_column_type
        self.full_name = full_name
        self.description = description
        self.key_column_alias = key_column_alias if key_column_alias else self.key_column

    def as_key(self, key_column_alias: str) -> TypedKey:
        """Rename the key alias. This is useful in derived features that depends on the same feature
            with different keys.
        """
        new_key = deepcopy(self)
        new_key.key_column_alias = key_column_alias
        return new_key


# passthrough/request feature do not need keys, as they are just a transformation defined on top of the request data
# They do not necessarily describe the value of keyed entity, e.g. dayofweek(timestamp) is transform on a request
# field without key
DUMMY_KEY = TypedKey(key_column="NOT_NEEDED",
                     key_column_type=ValueType.UNSPECIFIED,
                     full_name="feathr.dummy_typedkey",
                     description="A dummy typed key for passthrough/request feature.")


import glob
import importlib
import inspect
import itertools
import os
import re
import sys
import ast
import types
from graphlib import TopologicalSorter
from pathlib import Path
from tracemalloc import stop
from typing import Dict, List, Optional, Tuple, Union
from urllib.parse import urlparse
from time import sleep
from uuid import UUID

from azure.identity import DefaultAzureCredential
from jinja2 import Template
from loguru import logger
from pyapacheatlas.auth import ServicePrincipalAuthentication
from pyapacheatlas.auth.azcredential import AzCredentialWrapper
from pyapacheatlas.core import (AtlasClassification, AtlasEntity, AtlasProcess,
                                PurviewClient, TypeCategory)
from pyapacheatlas.core.typedef import (AtlasAttributeDef,
                                        AtlasRelationshipEndDef, Cardinality,
                                        EntityTypeDef, RelationshipTypeDef)
                            
from pyapacheatlas.core.util import GuidTracker,AtlasException
from pyhocon import ConfigFactory

from feathr.definition.dtype import *
from feathr.registry.registry_utils import *
from feathr.utils._file_utils import write_to_file
from feathr.definition.anchor import FeatureAnchor
from feathr.constants import *
from feathr.definition.feature import Feature, FeatureType,FeatureBase
from feathr.definition.feature_derivations import DerivedFeature
from feathr.definition.repo_definitions import RepoDefinitions
from feathr.definition.source import HdfsSource, InputContext, JdbcSource, Source
from feathr.definition.transformation import (ExpressionTransformation, Transformation,
                                   WindowAggTransformation)
from feathr.definition.typed_key import TypedKey
from feathr.registry.feature_registry import FeathrRegistry

from feathr.constants import *

def _to_snake(d, level: int = 0):
    """
    Convert `string`, `list[string]`, or all keys in a `dict` into snake case
    The maximum length of input string or list is 100, or it will be truncated before being processed, for dict, the exception will be thrown if it has more than 100 keys.
    the maximum nested level is 10, otherwise the exception will be thrown
    """
    if level >= 10:
        raise ValueError("Too many nested levels")
    if isinstance(d, str):
        d = d[:100]
        return re.sub(r'(?<!^)(?=[A-Z])', '_', d).lower()
    if isinstance(d, list):
        d = d[:100]
        return [_to_snake(i, level + 1) if isinstance(i, (dict, list)) else i for i in d]
    if len(d) > 100:
        raise ValueError("Dict has too many keys")
    return {_to_snake(a, level + 1): _to_snake(b, level + 1) if isinstance(b, (dict, list)) else b for a, b in d.items()}


class _PurviewRegistry(FeathrRegistry):
    """
    Initializes the feature registry, doing the following:
    - Use an DefaultAzureCredential() to communicate with Azure Purview
    - Initialize an Azure Purview Client
    - Initialize the GUID tracker, project name, etc.
    """
    def __init__(self, project_name: str, azure_purview_name: str, registry_delimiter: str, project_tags: Dict[str, str] = None, credential=None, config_path=None,):
        self.project_name = project_name
        self.registry_delimiter = registry_delimiter
        self.azure_purview_name = azure_purview_name
        self.project_tags = project_tags

        self.credential = DefaultAzureCredential(exclude_interactive_browser_credential=False) if credential is None else credential
        self.oauth = AzCredentialWrapper(credential=self.credential)
        self.purview_client = PurviewClient(
                account_name=self.azure_purview_name,
                authentication=self.oauth
            )   
        self.guid = GuidTracker(starting=-1000)
        self.entity_batch_queue = []

        # for searching in derived features by name
        self.global_feature_entity_dict = {}

    def _register_feathr_feature_types(self):
        """
        Register the feathr types if we haven't done so. Note that this only needs to be called once per provisioning
        a system. Basically this function registers all the feature type definition in a Atlas compatible system.
        """

        # Each feature is registered under a certain Feathr project. The project should what we refer to, however for backward compatibility, the type name would be `feathr_workspace`
        type_feathr_project = EntityTypeDef(
            name=TYPEDEF_FEATHR_PROJECT,
            attributeDefs=[
                # TODO: this should be called "anchors" rather than "anchor_features" to make it less confusing.
                AtlasAttributeDef(
                    name="anchor_features", typeName=TYPEDEF_ARRAY_ANCHOR, cardinality=Cardinality.SET),
                AtlasAttributeDef(
                    name="derived_features", typeName=TYPEDEF_ARRAY_DERIVED_FEATURE, cardinality=Cardinality.SET),
                AtlasAttributeDef(name="tags", typeName="map<string,string>",
                                  cardinality=Cardinality.SINGLE),
            ],
            superTypes=["DataSet"],

        )
        type_feathr_sources = EntityTypeDef(
            name=TYPEDEF_SOURCE,
            attributeDefs=[

                AtlasAttributeDef(
                    name="path", typeName="string", cardinality=Cardinality.SINGLE),
                AtlasAttributeDef(
                    name="url", typeName="string", cardinality=Cardinality.SINGLE),
                AtlasAttributeDef(
                    name="dbtable", typeName="string", cardinality=Cardinality.SINGLE),
                AtlasAttributeDef(
                    name="query", typeName="string", cardinality=Cardinality.SINGLE),
                AtlasAttributeDef(
                    name="auth", typeName="string", cardinality=Cardinality.SINGLE),
                AtlasAttributeDef(name="event_timestamp_column",
                                  typeName="string", cardinality=Cardinality.SINGLE),
                AtlasAttributeDef(name="timestamp_format",
                                  typeName="string", cardinality=Cardinality.SINGLE),
                AtlasAttributeDef(name="type", typeName="string",
                                  cardinality=Cardinality.SINGLE),
                AtlasAttributeDef(name="preprocessing", typeName="string",
                                  cardinality=Cardinality.SINGLE),
                AtlasAttributeDef(name="tags", typeName="map<string,string>",
                                  cardinality=Cardinality.SINGLE),
            ],
            superTypes=["DataSet"],
        )

        type_feathr_anchor_features = EntityTypeDef(
            name=TYPEDEF_ANCHOR_FEATURE,
            attributeDefs=[
                AtlasAttributeDef(name="type", typeName="string",
                                  cardinality=Cardinality.SINGLE),
                AtlasAttributeDef(name="key", typeName="array<map<string,string>>",
                                  cardinality=Cardinality.SET),
                AtlasAttributeDef(name="transformation", typeName="map<string,string>",
                                  cardinality=Cardinality.SINGLE),
                AtlasAttributeDef(name="tags", typeName="map<string,string>",
                                  cardinality=Cardinality.SINGLE),
            ],
            superTypes=["DataSet"],
        )

        type_feathr_derived_features = EntityTypeDef(
            name=TYPEDEF_DERIVED_FEATURE,
            attributeDefs=[
                AtlasAttributeDef(name="type", typeName="string",
                                  cardinality=Cardinality.SINGLE),

                AtlasAttributeDef(name="input_anchor_features", typeName=TYPEDEF_ARRAY_ANCHOR_FEATURE,
                                  cardinality=Cardinality.SET),
                AtlasAttributeDef(name="input_derived_features", typeName=TYPEDEF_ARRAY_DERIVED_FEATURE,
                                  cardinality=Cardinality.SET),
                AtlasAttributeDef(name="key", typeName="array<map<string,string>>",
                                  cardinality=Cardinality.SET),
                AtlasAttributeDef(name="transformation", typeName="map<string,string>",
                                  cardinality=Cardinality.SINGLE),
                AtlasAttributeDef(name="tags", typeName="map<string,string>",
                                  cardinality=Cardinality.SINGLE),
            ],
            superTypes=["DataSet"],
        )

        type_feathr_anchors = EntityTypeDef(
            name=TYPEDEF_ANCHOR,
            attributeDefs=[
                AtlasAttributeDef(
                    name="source", typeName=TYPEDEF_SOURCE, cardinality=Cardinality.SINGLE),
                AtlasAttributeDef(
                    name="features", typeName=TYPEDEF_ARRAY_ANCHOR_FEATURE, cardinality=Cardinality.SET),
                AtlasAttributeDef(name="tags", typeName="map<string,string>",
                                  cardinality=Cardinality.SINGLE),
            ],
            superTypes=["DataSet"],
        )

        def_result = self.purview_client.upload_typedefs(
            entityDefs=[type_feathr_anchor_features, type_feathr_anchors,
                        type_feathr_derived_features, type_feathr_sources, type_feathr_project],
            force_update=True)
        logger.info("Feathr Feature Type System Initialized.")

    def _parse_anchor_features(self, anchor: FeatureAnchor) -> List[AtlasEntity]:
        """
        This function will parse the anchor features and sources inside an anchor
        """

        anchor_feature_batch = []
        # annotate type
        anchor_feature: Feature

        for anchor_feature in anchor.features:
            key_list = []
            for individual_key in anchor_feature.key:
                key_dict = {"key_column": individual_key.key_column, "key_column_type": individual_key.key_column_type.value,
                            "full_name": individual_key.full_name, "description": individual_key.description, "key_column_alias": individual_key.key_column_alias}
                key_list.append(key_dict)

            # define a dict to save all the transformation schema
            transform_dict = {}
            if isinstance(anchor_feature.transform, ExpressionTransformation):
                transform_dict = {"transform_expr": anchor_feature.transform.expr}
            elif isinstance(anchor_feature.transform, WindowAggTransformation):
                transform_dict = {
                    "def_expr": anchor_feature.transform.def_expr,
                    "agg_func": anchor_feature.transform.agg_func,
                    "window": anchor_feature.transform.window,
                    "group_by": anchor_feature.transform.group_by,
                    "filter": anchor_feature.transform.filter,
                    "limit": anchor_feature.transform.limit,
                }

            anchor_feature_entity = AtlasEntity(
                name=anchor_feature.name,
                qualified_name=self.project_name + self.registry_delimiter +
                anchor.name + self.registry_delimiter + anchor_feature.name,
                attributes={
                    "type": anchor_feature.feature_type.to_feature_config(),
                    "key": key_list,
                    "transformation": transform_dict,
                    "tags": anchor_feature.registry_tags,
                },
                typeName=TYPEDEF_ANCHOR_FEATURE,
                guid=self.guid.get_guid(),
            )
            self.entity_batch_queue.append(anchor_feature_entity)
            # add the entity to a dict that can search by name
            self.global_feature_entity_dict[anchor_feature.name] = anchor_feature_entity
            anchor_feature_batch.append(anchor_feature_entity)
        return anchor_feature_batch

    def _parse_anchors(self, anchor_list: List[FeatureAnchor]) -> List[AtlasEntity]:
        """
        parse content of an anchor
        """
        anchors_batch = []
        for anchor in anchor_list:
            # First, parse all the features in this anchor
            anchor_feature_entities = self._parse_anchor_features(anchor)
            # then parse the source of that anchor
            source_entity = self._parse_source(anchor.source)
            anchor_fully_qualified_name  = self.project_name+self.registry_delimiter+anchor.name
            original_id = self.get_feature_id(anchor_fully_qualified_name, type=TYPEDEF_ANCHOR )
            original_anchor = self.get_feature_by_guid(original_id) if original_id else None
            merged_elements = self._merge_anchor(original_anchor,anchor_feature_entities)
            anchor_entity = AtlasEntity(
                name=anchor.name,
                qualified_name=anchor_fully_qualified_name ,
                attributes={
                    "source": source_entity.to_json(minimum=True),
                    "features": merged_elements,
                    "tags": anchor.registry_tags
                },
                typeName=TYPEDEF_ANCHOR,
                guid=self.guid.get_guid(),
            )
            # add feature lineage between anchor and feature
            for anchor_feature_entity in anchor_feature_entities:
                lineage = AtlasProcess(
                    name=anchor_feature_entity.name + " to " + anchor.name,
                    typeName="Process",
                    qualified_name=self.registry_delimiter + "PROCESS" + self.registry_delimiter + self.project_name +
                    self.registry_delimiter + anchor.name + self.registry_delimiter +
                    anchor_feature_entity.name,
                    inputs=[anchor_feature_entity],
                    outputs=[anchor_entity],
                    guid=self.guid.get_guid(),
                )
                self.entity_batch_queue.append(lineage)

            # add lineage between anchor and source
            anchor_source_lineage = AtlasProcess(
                name=source_entity.name + " to " + anchor.name,
                typeName="Process",
                qualified_name=self.registry_delimiter + "PROCESS" + self.registry_delimiter + self.project_name +
                self.registry_delimiter + anchor.name + self.registry_delimiter +
                source_entity.name,
                inputs=[source_entity],
                outputs=[anchor_entity],
                guid=self.guid.get_guid(),
            )
            self.entity_batch_queue.append(anchor_source_lineage)

            anchors_batch.append(anchor_entity)
        return anchors_batch

    def _merge_anchor(self,original_anchor:Dict, new_anchor:Dict)->List[Dict[str,any]]:
        '''
        Merge the new anchors defined locally with the anchors that is defined in the centralized registry.
        '''
        # TODO: This will serve as a quick fix, full fix will work with MVCC, and is in progress.
        new_anchor_json_repr = [s.to_json(minimum=True) for s in new_anchor]
        if not original_anchor:
            # if the anchor is not present on the registry, return json representation of the locally defined anchor.
            # sample : [{'guid':'GUID_OF_ANCHOR','typeName':'','qualifiedName':'QUALIFIED_NAME'}
            return new_anchor_json_repr
        else:
            original_anchor_elements = [x for x in original_anchor['entity']['attributes']['features']]
            transformed_original_elements = {
                x['uniqueAttributes']['qualifiedName']:
                {
                    'guid':x['guid'],
                    'typeName':x['typeName'],
                    'qualifiedName':x['uniqueAttributes']['qualifiedName']
                }
                for x in original_anchor_elements}
            for elem in new_anchor_json_repr:
                transformed_original_elements.setdefault(elem['qualifiedName'],elem)
            return list(transformed_original_elements.values())
            
    def _parse_source(self, source: Union[Source, HdfsSource, JdbcSource]) -> AtlasEntity:
        """
        parse the input sources
        """
        input_context = False
        if isinstance(source, InputContext):
            input_context = True
        
        # only set preprocessing if it's available in the object and is not None
        if 'preprocessing' in dir(source) and source.preprocessing is not None:
            preprocessing_func = inspect.getsource(source.preprocessing)
        else:
            preprocessing_func = None

        attrs = {}
        if isinstance(source, JdbcSource):
            {
                "type": INPUT_CONTEXT if input_context else urlparse(source.path).scheme,
                "url": INPUT_CONTEXT if input_context else source.url,
                "timestamp_format": source.timestamp_format,
                "event_timestamp_column": source.event_timestamp_column,
                "tags": source.registry_tags,
                "preprocessing": preprocessing_func  # store the UDF as a string
            }
            if source.auth is not None:
                attrs["auth"] = source.auth
            if source.dbtable is not None:
                attrs["dbtable"] = source.dbtable
            if source.query is not None:
                attrs["query"] = source.query
        else:
            attrs = {
                "type": INPUT_CONTEXT if input_context else urlparse(source.path).scheme,
                "path": INPUT_CONTEXT if input_context else source.path,
                "timestamp_format": source.timestamp_format,
                "event_timestamp_column": source.event_timestamp_column,
                "tags": source.registry_tags,
                "preprocessing": preprocessing_func  # store the UDF as a string
            }
        source_entity = AtlasEntity(
            name=source.name,
            qualified_name=self.project_name + self.registry_delimiter + source.name,
            attributes=attrs,
            typeName=TYPEDEF_SOURCE,
            guid=self.guid.get_guid(),
        )
        self.entity_batch_queue.append(source_entity)
        return source_entity

    def _add_all_derived_features(self, derived_features: List[DerivedFeature], ts:TopologicalSorter ) -> None:
        """iterate thru all the dependencies of the derived feature and return a derived feature list in a topological sorted way (the result list only has derived features, without their anchor features)

        Args:
            derived_features (List[DerivedFeature]): input derive feature list
            ts (TopologicalSorter): a topological sorter by python

        Returns:
            None. The topological sorter will maintain a static topological sorted order.
        """
        # return if the list is empty
        if derived_features is None:
            return

        for derived_feature in derived_features:
            # make sure the input is derived feature
            if isinstance(derived_feature, DerivedFeature):
                # add this derived feature in the topological sort graph without any predecessors
                # since regardless we need it
                ts.add(derived_feature)
                for input_feature in derived_feature.input_features:
                    if isinstance(input_feature, DerivedFeature):
                        # `input_feature` is predecessor of `derived_feature`
                        ts.add(derived_feature, input_feature)
                        # if any of the input feature is a derived feature, have this recursive call
                        # use this for code simplicity. 
                        # if the amount of features is huge, consider only add the derived features into the function call
                        self._add_all_derived_features(input_feature.input_features, ts)

                        


    def _parse_derived_features(self, derived_features: List[DerivedFeature]) -> List[AtlasEntity]:
        """parse derived feature

        Args:
            derived_features (List[DerivedFeature]): derived feature in a list fashion. This function will handle the derived feature dependencies using topological sort to ensure that the corresponding features can be parsed correctly

        Returns:
            List[AtlasEntity]: list of parsed Atlas entities for each of the derived feature
        """
        derivation_entities = []
        ts = TopologicalSorter()

        self._add_all_derived_features(derived_features, ts)
        # topological sort the derived features to make sure that we can correctly refer to them later in the registry
        toposorted_derived_feature_list: List[DerivedFeature] = list(ts.static_order())
        
        for derived_feature in toposorted_derived_feature_list:
            # get the corresponding Atlas entity by searching feature name
            # Since this list is topological sorted, so you can always find the corresponding name
            input_feature_entity_list: List[AtlasEntity] = [
                self.global_feature_entity_dict[f.name] for f in derived_feature.input_features]
            key_list = []
            for individual_key in derived_feature.key:
                key_dict = {"key_column": individual_key.key_column, "key_column_type": individual_key.key_column_type.value,
                            "full_name": individual_key.full_name, "description": individual_key.description, "key_column_alias": individual_key.key_column_alias}
                key_list.append(key_dict)

            # define a dict to save all the transformation schema
            transform_dict = {}
            if isinstance(derived_feature.transform, ExpressionTransformation):
                transform_dict = {"transform_expr": derived_feature.transform.expr}
            elif isinstance(derived_feature.transform, WindowAggTransformation):
                transform_dict = {
                    "def_expr": derived_feature.transform.def_expr,
                    "agg_func": derived_feature.transform.agg_func,
                    "window": derived_feature.transform.window,
                    "group_by": derived_feature.transform.group_by,
                    "filter": derived_feature.transform.filter,
                    "limit": derived_feature.transform.limit,
                }
            
            derived_feature_entity = AtlasEntity(
                name=derived_feature.name,
                qualified_name=self.project_name +
                self.registry_delimiter + derived_feature.name,
                attributes={
                    "type": derived_feature.feature_type.to_feature_config(),
                    "key": key_list,
                    "input_anchor_features": [f.to_json(minimum=True) for f in input_feature_entity_list if f.typeName==TYPEDEF_ANCHOR_FEATURE],
                    "input_derived_features": [f.to_json(minimum=True) for f in input_feature_entity_list if f.typeName==TYPEDEF_DERIVED_FEATURE],
                    "transformation": transform_dict,
                    "tags": derived_feature.registry_tags,

                },
                typeName=TYPEDEF_DERIVED_FEATURE,
                guid=self.guid.get_guid(),
            )

            # Add the feature entity in the global dict so that it can be referenced further. 
            self.global_feature_entity_dict[derived_feature.name] = derived_feature_entity

            for input_feature_entity in input_feature_entity_list:
                # add lineage between anchor feature and derived feature
                derived_feature_feature_lineage = AtlasProcess(
                    name=input_feature_entity.name + " to " + derived_feature.name,
                    typeName="Process",
                    qualified_name=self.registry_delimiter + "PROCESS" + self.registry_delimiter + self.project_name +
                    self.registry_delimiter + derived_feature.name + self.registry_delimiter +
                    input_feature_entity.name,
                    inputs=[input_feature_entity],
                    outputs=[derived_feature_entity],
                    guid=self.guid.get_guid(),
                )
                self.entity_batch_queue.append(derived_feature_feature_lineage)

            self.entity_batch_queue.append(derived_feature_entity)
            derivation_entities.append(derived_feature_entity)
        return derivation_entities

    def _parse_features_from_context(self, workspace_path: str, anchor_list, derived_feature_list):
        """
        Read feature content from python objects (which is provided in the context)
        """
        # define it here to make sure the variable is accessible
        anchor_entities = derived_feature_entities = []

        # parse all the anchors
        if anchor_list:
            anchor_entities = self._parse_anchors(anchor_list)

        project_attributes = {"anchor_features": [
            s.to_json(minimum=True) for s in anchor_entities], "tags": self.project_tags}
        # add derived feature if it's there
        if derived_feature_list:
            derived_feature_entities = self._parse_derived_features(
                derived_feature_list)
            project_attributes["derived_features"] = [
                s.to_json(minimum=True) for s in derived_feature_entities]

        # define project in Atlas entity
        feathr_project_entity = AtlasEntity(
            name=self.project_name,
            qualified_name=self.project_name,
            attributes=project_attributes,
            typeName=TYPEDEF_FEATHR_PROJECT,
            guid=self.guid.get_guid(),
        )

        # add lineage from anchor to project
        for individual_anchor_entity in anchor_entities:

            lineage_process = AtlasProcess(
                name=individual_anchor_entity.name + " to " + self.project_name,
                typeName="Process",
                # fqdn: PROCESS+PROJECT_NAME+ANCHOR_NAME
                qualified_name=self.registry_delimiter + "PROCESS" + self.registry_delimiter + self.project_name +
                self.registry_delimiter + individual_anchor_entity.name,
                inputs=[individual_anchor_entity],
                outputs=[feathr_project_entity],
                guid=self.guid.get_guid(),
            )
            self.entity_batch_queue.append(lineage_process)

        # add lineage from derivation to project
        for derived_feature_entity in derived_feature_entities:
            lineage_process = AtlasProcess(
                name=derived_feature_entity.name + " to " + self.project_name,
                typeName="Process",
                # fqdn: PROCESS+PROJECT_NAME+DERIVATION_NAME
                qualified_name=self.registry_delimiter + "PROCESS" + self.registry_delimiter + self.project_name +
                self.registry_delimiter + derived_feature_entity.name,
                inputs=[derived_feature_entity],
                outputs=[feathr_project_entity],
                guid=self.guid.get_guid(),
            )
            self.entity_batch_queue.append(lineage_process)
            
        self.entity_batch_queue.append(feathr_project_entity)
        self.entity_batch_queue.extend(anchor_entities)
        self.entity_batch_queue.extend(derived_feature_entities)

    @classmethod
    def _get_py_files(self, path: Path) -> List[Path]:
        """Get all Python files under path recursively, excluding __init__.py"""
        py_files = []
        for item in path.glob('**/*.py'):
            if "__init__.py" != item.name:
                py_files.append(item)
        return py_files

    @classmethod
    def _convert_to_module_path(self, path: Path, workspace_path: Path) -> str:
        """Convert a Python file path to its module path so that we can import it later"""
        prefix = os.path.commonprefix(
            [path.resolve(), workspace_path.resolve()])
        resolved_path = str(path.resolve())
        module_path = resolved_path[len(prefix): -len(".py")]
        # Convert features under nested folder to module name
        # e.g. /path/to/pyfile will become path.to.pyfile
        return (
            module_path
            .lstrip('/')
            .replace("/", ".")
        )

    @classmethod
    def _extract_features_from_context(self, anchor_list, derived_feature_list, result_path: Path) -> RepoDefinitions:
        """Collect feature definitions from the context instead of python files"""
        definitions = RepoDefinitions(
            sources=set(),
            features=set(),
            transformations=set(),
            feature_anchors=set(),
            derived_features=set()
        )
        for derived_feature in derived_feature_list:
            if isinstance(derived_feature, DerivedFeature):
                definitions.derived_features.add(derived_feature)
                definitions.transformations.add(
                    vars(derived_feature)["transform"])
            else:
                raise RuntimeError(
                    "Object cannot be parsed. `derived_feature_list` should be a list of `DerivedFeature`.")

        for anchor in anchor_list:
            # obj is `FeatureAnchor`
            definitions.feature_anchors.add(anchor)
            # add the source section of this `FeatureAnchor` object
            definitions.sources.add(vars(anchor)['source'])
            for feature in vars(anchor)['features']:
                # get the transformation object from `Feature` or `DerivedFeature`
                if isinstance(feature, Feature):
                    # feature is of type `Feature`
                    definitions.features.add(feature)
                    definitions.transformations.add(vars(feature)["transform"])
                else:
                    raise RuntimeError("Object cannot be parsed.")

        return definitions

    @classmethod
    def _extract_features(self, workspace_path: Path) -> RepoDefinitions:
        """Collect feature definitions from the python file, convert them into feature config and save them locally"""
        os.chdir(workspace_path)
        # Add workspace path to system path so that we can load features defined in Python via import_module
        sys.path.append(str(workspace_path))
        definitions = RepoDefinitions(
            sources=set(),
            features=set(),
            transformations=set(),
            feature_anchors=set(),
            derived_features=set()
        )
        for py_file in self._get_py_files(workspace_path):
            module_path = self._convert_to_module_path(py_file, workspace_path)
            module = importlib.import_module(module_path)
            for attr_name in dir(module):
                obj = getattr(module, attr_name)
                if isinstance(obj, Source):
                    definitions.sources.add(obj)
                elif isinstance(obj, Feature):
                    definitions.features.add(obj)
                elif isinstance(obj, DerivedFeature):
                    definitions.derived_features.add(obj)
                elif isinstance(obj, FeatureAnchor):
                    definitions.feature_anchors.add(obj)
                elif isinstance(obj, Transformation):
                    definitions.transformations.add(obj)
        return definitions

    @classmethod
    def save_to_feature_config(self, workspace_path: Path, config_save_dir: Path):
        """Save feature definition within the workspace into HOCON feature config files"""
        repo_definitions = self._extract_features(workspace_path)
        self._save_request_feature_config(repo_definitions, config_save_dir)
        self._save_anchored_feature_config(repo_definitions, config_save_dir)
        self._save_derived_feature_config(repo_definitions, config_save_dir)

    @classmethod
    def save_to_feature_config_from_context(self, anchor_list, derived_feature_list, local_workspace_dir: Path):
        """Save feature definition within the workspace into HOCON feature config files from current context, rather than reading from python files"""
        repo_definitions = self._extract_features_from_context(
            anchor_list, derived_feature_list, local_workspace_dir)
        self._save_request_feature_config(repo_definitions, local_workspace_dir)
        self._save_anchored_feature_config(repo_definitions, local_workspace_dir)
        self._save_derived_feature_config(repo_definitions, local_workspace_dir)

    @classmethod
    def _save_request_feature_config(self, repo_definitions: RepoDefinitions, local_workspace_dir="./"):
        config_file_name = "feature_conf/auto_generated_request_features.conf"
        tm = Template(
            """
// THIS FILE IS AUTO GENERATED. PLEASE DO NOT EDIT.
anchors: {
    {% for anchor in feature_anchors %}
        {% if anchor.source.name == "PASSTHROUGH" %}
            {{anchor.to_feature_config()}}
        {% endif %}
    {% endfor %}
}
"""
        )

        request_feature_configs = tm.render(
            feature_anchors=repo_definitions.feature_anchors)
        config_file_path = os.path.join(local_workspace_dir, config_file_name)
        write_to_file(content=request_feature_configs,
                      full_file_name=config_file_path)

    @classmethod
    def _save_anchored_feature_config(self, repo_definitions: RepoDefinitions, local_workspace_dir="./"):
        config_file_name = "feature_conf/auto_generated_anchored_features.conf"
        tm = Template(
            """
// THIS FILE IS AUTO GENERATED. PLEASE DO NOT EDIT.
anchors: {
    {% for anchor in feature_anchors %}
        {% if not anchor.source.name == "PASSTHROUGH" %}
            {{anchor.to_feature_config()}}
        {% endif %}
    {% endfor %}
}

sources: {
    {% for source in sources%}
        {% if not source.name == "PASSTHROUGH" %}
            {{source.to_feature_config()}}
        {% endif %}
    {% endfor %}
}
"""
        )
        anchored_feature_configs = tm.render(feature_anchors=repo_definitions.feature_anchors,
                                             sources=repo_definitions.sources)
        config_file_path = os.path.join(local_workspace_dir, config_file_name)
        write_to_file(content=anchored_feature_configs,
                      full_file_name=config_file_path)

    @classmethod
    def _save_derived_feature_config(self, repo_definitions: RepoDefinitions, local_workspace_dir="./"):
        config_file_name = "feature_conf/auto_generated_derived_features.conf"
        tm = Template(
            """
anchors: {}
derivations: {
    {% for derived_feature in derived_features %}
        {{derived_feature.to_feature_config()}}
    {% endfor %}
}
"""
        )
        derived_feature_configs = tm.render(
            derived_features=repo_definitions.derived_features)
        config_file_path = os.path.join(local_workspace_dir, config_file_name)
        write_to_file(content=derived_feature_configs,
                      full_file_name=config_file_path)
                      
    def _create_project(self) -> UUID:
        ''' 
        create a project entity
        ''' 
        predefined_guid = self.guid.get_guid()
        feathr_project_entity = AtlasEntity(
            name=self.project_name,
            qualified_name=self.project_name,
            typeName=TYPEDEF_FEATHR_PROJECT,
            guid=predefined_guid)
        guid = self.upload_single_entity_to_purview(feathr_project_entity)
        return guid

    def upload_single_entity_to_purview(self,entity:Union[AtlasEntity,AtlasProcess]):
        '''
        Upload a single entity to purview, could be a process entity or AtlasEntity. 
        Since this is used for migration existing project, ignore Atlas PreconditionFail (412)
        If the entity already exists, return the existing entity's GUID.
        Otherwise, return the new entity GUID.
        The entity itself will also be modified, fill the GUID with real GUID in Purview.
        In order to avoid having concurrency issue, and provide clear guidance, this method only allows entity uploading once at a time.
        '''
        try:
            """
            Try to find existing entity/process first, if found, return the existing entity's GUID
            """
            id = self.get_entity_id(entity.qualifiedName)
            response =  self.purview_client.get_entity(id)['entities'][0]
            j = entity.to_json()
            if j["typeName"] == response["typeName"]:
                if j["typeName"] == "Process":
                    if response["attributes"]["qualifiedName"] != j["attributes"]["qualifiedName"]:
                        raise RuntimeError("The requested entity %s conflicts with the existing entity in PurView" % j["attributes"]["qualifiedName"])
                else:
                    if "type" in response['attributes'] and response["typeName"] in (TYPEDEF_ANCHOR_FEATURE, TYPEDEF_DERIVED_FEATURE):
                        conf = ConfigFactory.parse_string(response['attributes']['type'])
                        response['attributes']['type'] = dict(conf)
                    keys = set([_to_snake(key) for key in j["attributes"].keys()]) - set(["qualified_name"])
                    keys.add("qualifiedName")
                    for k in keys:
                        if response["attributes"][k] != j["attributes"][k]:
                            raise RuntimeError("The requested entity %s conflicts with the existing entity in PurView" % j["attributes"]["qualifiedName"])
                return response["guid"]
            else:
                raise RuntimeError("The requested entity %s conflicts with the existing entity in PurView" % j["attributes"]["qualifiedName"])
        except AtlasException as e:
            pass
        
        try:
            entity.lastModifiedTS="0"
            result = self.purview_client.upload_entities([entity])
            entity.guid = result['guidAssignments'][entity.guid]
            print(f"Successfully created {entity.typeName} -- {entity.qualifiedName}")
        except AtlasException as e:
            if "PreConditionCheckFailed" in e.args[0]:
                entity.guid = self.purview_client.get_entity(qualifiedName=entity.qualifiedName,typeName = entity.typeName)['entities'][0]['guid']
                print(f"Found existing entity  {entity.guid}, {entity.typeName} -- {entity.qualifiedName}")
        return UUID(entity.guid)
    
    def _generate_relation_pairs(self, from_entity:dict, to_entity:dict, relation_type):
        type_lookup = {RELATION_CONTAINS: RELATION_BELONGSTO, RELATION_CONSUMES: RELATION_PRODUCES}

        forward_relation =  AtlasProcess(
            name=str(from_entity["guid"]) + " to " + str(to_entity["guid"]),
            typeName="Process",
            qualified_name=self.registry_delimiter.join(
                [relation_type,str(from_entity["guid"]), str(to_entity["guid"])]),
            inputs=[self.to_min_repr(from_entity)],
            outputs=[self.to_min_repr(to_entity)],
            guid=self.guid.get_guid())
        
        backward_relation = AtlasProcess(
            name=str(to_entity["guid"]) + " to " + str(from_entity["guid"]),
            typeName="Process",
            qualified_name=self.registry_delimiter.join(
                [type_lookup[relation_type], str(to_entity["guid"]), str(from_entity["guid"])]),
            inputs=[self.to_min_repr(to_entity)],
            outputs=[self.to_min_repr(from_entity)],
            guid=self.guid.get_guid())
        return [forward_relation,backward_relation]
    
    def to_min_repr(self,entity:dict) -> dict:
        return {
            'qualifiedName':entity['attributes']["qualifiedName"],
            'guid':str(entity["guid"]),
            'typeName':str(entity['typeName']),
        }

    def _create_source(self, s: Source) -> UUID:
        '''
        create a data source under a project.
        this will create the data source entity, together with the relation entity
        '''
        project_entity = self.purview_client.get_entity(qualifiedName=self.project_name,typeName=TYPEDEF_FEATHR_PROJECT)['entities'][0]
        attrs = source_to_def(s)
        qualified_name = self.registry_delimiter.join([project_entity['attributes']['qualifiedName'],attrs['name']])
        source_entity = AtlasEntity(
            name=attrs['name'],
            qualified_name=qualified_name,
            attributes= {k:v for k,v in attrs.items() if k !="name"},
            typeName=TYPEDEF_SOURCE,
            guid=self.guid.get_guid(),
        )
        source_id = self.upload_single_entity_to_purview(source_entity)
        
        # change from AtlasEntity to Entity
        source_entity = self.purview_client.get_entity(source_id)['entities'][0]

        # Project contains source, source belongs to project
        project_contains_source_relation = self._generate_relation_pairs(
            project_entity, source_entity, RELATION_CONTAINS)
        [self.upload_single_entity_to_purview(x) for x in project_contains_source_relation]
        return source_id

    def _create_anchor(self, s: FeatureAnchor) -> UUID:
        '''
        Create anchor under project ,and based on the data source
        This will also create two relation pairs 
        '''
        project_entity = self.purview_client.get_entity(qualifiedName=self.project_name,typeName=TYPEDEF_FEATHR_PROJECT)['entities'][0]
        source_entity = self.purview_client.get_entity(qualifiedName=self.registry_delimiter.join([self.project_name,s.source.name]),typeName=TYPEDEF_SOURCE)['entities'][0]
        attrs = anchor_to_def(s)
        qualified_name = self.registry_delimiter.join([self.project_name,attrs['name']])
        anchor_entity = AtlasEntity(
            name=s.name,
            qualified_name=qualified_name,
            attributes= {k:v for k,v in attrs.items() if k not in ['name','qualifiedName']},
            typeName=TYPEDEF_ANCHOR,
            guid=self.guid.get_guid(),
        )

        anchor_id = self.upload_single_entity_to_purview(anchor_entity)

        # change from AtlasEntity to Entity
        anchor_entity = self.purview_client.get_entity(anchor_id)['entities'][0]

        
        # project contains anchor, anchor belongs to project.
        project_contains_anchor_relation = self._generate_relation_pairs(
            project_entity, anchor_entity, RELATION_CONTAINS)
        anchor_consumes_source_relation = self._generate_relation_pairs(
            anchor_entity,source_entity, RELATION_CONSUMES)
        [self.upload_single_entity_to_purview(x) for x in project_contains_anchor_relation + anchor_consumes_source_relation]

        return anchor_id

    def _create_anchor_feature(self, anchor_id: str, source:Source,s: Feature) -> UUID:
        '''
        Create anchor feature under anchor. 
        This will also create three relation pairs
        '''
        project_entity = self.purview_client.get_entity(qualifiedName=self.project_name,typeName=TYPEDEF_FEATHR_PROJECT)['entities'][0]
        anchor_entity = self.purview_client.get_entity(anchor_id)['entities'][0]
        source_entity = self.purview_client.get_entity(qualifiedName=self.registry_delimiter.join([self.project_name,source.name]),typeName=TYPEDEF_SOURCE)['entities'][0]

        attrs = feature_to_def(s)
        attrs['type'] = attrs['featureType']
        qualified_name = self.registry_delimiter.join([self.project_name,
                                                       anchor_entity["attributes"]["name"],
                                                       attrs['name']])

        anchor_feature_entity = AtlasEntity(
            name=s.name,
            qualified_name=qualified_name,
            attributes= {k:v for k,v in attrs.items() if k not in ['name','qualifiedName']},
            typeName=TYPEDEF_ANCHOR_FEATURE,
            guid=self.guid.get_guid())
        anchor_feature_id = self.upload_single_entity_to_purview(anchor_feature_entity)

        # change from AtlasEntity to Entity
        anchor_feature_entity = self.purview_client.get_entity(anchor_feature_id)['entities'][0]

        # Project contains AnchorFeature, AnchorFeature belongs to Project
        project_contains_feature_relation = self._generate_relation_pairs(
            project_entity, anchor_feature_entity, RELATION_CONTAINS)
        anchor_contains_feature_relation = self._generate_relation_pairs(
            anchor_entity, anchor_feature_entity, RELATION_CONTAINS)
        feature_consumes_source_relation = self._generate_relation_pairs(
            anchor_feature_entity, source_entity, RELATION_CONSUMES)

        [self.upload_single_entity_to_purview(x) for x in  
        project_contains_feature_relation
        + anchor_contains_feature_relation
        + feature_consumes_source_relation]
        
        return anchor_feature_id

    def _create_derived_feature(self, s: DerivedFeature) -> UUID:
        '''
        Create DerivedFeature.
        This will also create multiple relations.
        '''
        input_features = [self.purview_client.get_entity(x._registry_id)['entities'][0] for x in s.input_features]
        attrs = derived_feature_to_def(s)
        attrs['type'] = attrs['featureType']

        project_entity = self.purview_client.get_entity(qualifiedName=self.project_name,typeName=TYPEDEF_FEATHR_PROJECT)['entities'][0]
        qualified_name = self.registry_delimiter.join([self.project_name,attrs['name']])
        derived_feature_entity = AtlasEntity(
            name=s.name,
            qualified_name=qualified_name,
            attributes={k:v for k,v in attrs.items() if k not in ['name','qualifiedName']},
            typeName=TYPEDEF_DERIVED_FEATURE,
            guid=self.guid.get_guid())
        derived_feature_id = self.upload_single_entity_to_purview(derived_feature_entity)
        
        # change from AtlasEntity to Entity
        derived_feature_entity = self.purview_client.get_entity(derived_feature_id)['entities'][0]

        # Project contains DerivedFeature, DerivedFeature belongs to Project.
        feature_project_contain_belong_pairs = self._generate_relation_pairs(
            project_entity, derived_feature_entity, RELATION_CONTAINS)

        consume_produce_pairs = []
        # Each input feature produces DerivedFeature, DerivedFeatures consumes from each input feature.
        for input_feature in input_features:
            consume_produce_pairs += self._generate_relation_pairs(
                    derived_feature_entity, input_feature,RELATION_CONSUMES)
        
        [self.upload_single_entity_to_purview(x) for x in  
        feature_project_contain_belong_pairs
            + consume_produce_pairs]
        
        return derived_feature_id

    def register_features(self, workspace_path: Optional[Path] = None, from_context: bool = True, anchor_list:List[FeatureAnchor]=[], derived_feature_list:List[DerivedFeature]=[]):
        """Register Features for the specified workspace. 

        Args:
            workspace_path (str, optional): path to a workspace. Defaults to None.
            from_context: whether the feature is from context (i.e. end users has to callFeathrClient.build_features()) or the feature is from a pre-built config file. Currently Feathr only supports register features from context.
            anchor_list: The anchor list after feature build
            derived_feature_list: the derived feature list after feature build
        """

        if not from_context:
            raise RuntimeError("Currently Feathr only supports registering features from context (i.e. you must call FeathrClient.build_features() before calling this function).")
                # Before starting, create the project
        self._register_feathr_feature_types()

        self.project_id = self._create_project()

        for anchor in anchor_list:
            source = anchor.source
            # 1. Create Source on the registry
            # We always re-create INPUT_CONTEXT as lots of existing codes reuse the singleton in different projects
            if (source.name == INPUT_CONTEXT) or (not hasattr(source, "_registry_id")):
                source._registry_id = self._create_source(source)
            # 2. Create Anchor on the registry
            if not hasattr(anchor, "_registry_id"):
                anchor._registry_id = self._create_anchor(anchor)
            # 3. Create all features on the registry
            for feature in anchor.features:
                if not hasattr(feature, "_registry_id"):
                    feature._registry_id = self._create_anchor_feature(
                        anchor._registry_id, anchor.source,feature)
        # 4. Create all derived features on the registry
        for df in topological_sort(derived_feature_list):
            if not hasattr(df, "_registry_id"):
                df._registry_id = self._create_derived_feature(df)
        logger.info(
            "Finished registering features.")


    def _purge_feathr_registry(self):
        """
        Delete all the feathr related entities and type definitions in feathr registry. For internal use only
        """
        self._delete_all_feathr_entities()
        self._delete_all_feathr_types()


    def _delete_all_feathr_types(self):
        """
        Delete all the corresponding type definitions for feathr registry. For internal use only
        """
        typedefs = self.purview_client.get_all_typedefs()

        relationshipdef_list=[]
        for relationshipdef in typedefs['relationshipDefs']:
            if "feathr" in relationshipdef['name']:
                relationshipdef_list.append(relationshipdef)
        self.purview_client.delete_typedefs(relationshipDefs=relationshipdef_list)
        

        entitydef_list=[]
        for typedef in typedefs['entityDefs']:
            if "feathr" in typedef['name']:
                entitydef_list.append(typedef )
        self.purview_client.delete_typedefs(entityDefs=entitydef_list)

        logger.info("Deleted all the Feathr related definitions.")

    def _delete_all_feathr_entities(self):
        """
        Delete all the corresponding entity for feathr registry. For internal use only

        :param guid: The guid or guids you want to remove.
        """
        # should not be large than this, otherwise the backend might throw out error
        batch_delete_size = 100

        # use the `query` API so that it can return immediately (don't use the search_entity API as it will try to return all the results in a single request)

        while True:
            result = self.purview_client.discovery.query(
                "feathr", limit=batch_delete_size)
            logger.info("Total number of entities:",result['@search.count'] )

            # if no results, break:
            if result['@search.count']  == 0:
                break
            entities = result['value']
            guid_list = [entity["id"] for entity in entities]
            self.purview_client.delete_entity(guid=guid_list)
            logger.info("{} feathr entities deleted", batch_delete_size)
            # sleep here, otherwise backend might throttle
            # process the next batch after sleep
            sleep(1)
    
    @classmethod
    def _get_registry_client(self):
        """
        Return a client object and users can operate more on it (like doing search)
        """
        return self.purview_client

    def list_registered_features(self, project_name: str, limit=1000, starting_offset=0) -> List[Dict[str,str]]:
        """
        List all the already registered features. If project_name is not provided or is None, it will return all the
        registered features; otherwise it will only return only features under this project
        """

        feature_list = []

        if not project_name:
            raise RuntimeError("project_name must be specified.")

        # get the corresponding features belongs to a certain project.
        # note that we need to use "startswith" to filter out the features that don't belong to this project.
        # see syntax here: https://docs.microsoft.com/en-us/rest/api/purview/catalogdataplane/discovery/query#discovery_query_andornested
        query_filter = {
            "and": [
                {
                    "or":
                    [
                        {"entityType": TYPEDEF_DERIVED_FEATURE},
                        {"entityType": TYPEDEF_ANCHOR_FEATURE}
                    ]
                },
                {
                    "attributeName": "qualifiedName",
                    "operator": "startswith",
                    "attributeValue": project_name + self.registry_delimiter
                }
            ]
        }
        result = self.purview_client.discovery.query(filter=query_filter)

        entities = result['value']
        # entities = self.purview_client.discovery.search_entities(query = None, search_filter=query_filter, limit=limit)

        for entity in entities:
            feature_list.append({"name":entity["name"],'id':entity['id'],"qualifiedName":entity['qualifiedName']})

        return feature_list
   
    def get_feature_by_fqdn_type(self, qualifiedName, typeName):
        """
        Get a single feature by it's QualifiedName and Type
        Returns the feature else throws an AtlasException with 400 error code
        """
        response = self.purview_client.get_entity(qualifiedName=qualifiedName, typeName=typeName)
        entities = response.get('entities')
        for entity in entities:
            if entity.get('typeName') == typeName and entity.get('attributes').get('qualifiedName') == qualifiedName: 
                return entity
       
    def get_feature_by_fqdn(self, qualifiedName):
        """
        Get feature by qualifiedName
        Returns the feature else throws an AtlasException with 400 error code
        """        
        id = self.get_feature_id(qualifiedName)
        return self.get_feature_by_guid(id)
    
    def get_feature_by_guid(self, guid):
        """
        Get a single feature by it's GUID
        Returns the feature else throws an AtlasException with 400 error code
        """ 
        response = self.purview_client.get_single_entity(guid=guid)
        return response
    
    def get_feature_lineage(self, guid):
        """
        Get feature's lineage by it's GUID
        Returns the feature else throws an AtlasException with 400 error code
        """
        return self.purview_client.get_entity_lineage(guid=guid)

    def get_feature_id(self, qualifiedName, type: str):
        """
        Get guid of a feature given its qualifiedName
        """        
        # the search term should be full qualified name
        # TODO: need to update the calling functions to add `type` field to make it more performant
        # purview_client.get_entity(qualifiedName=qualifiedName) might not work here since it requires an additonal typeName parameter
        # Currently still use the `query` API to get the result in a "full name match" way.
        # self.purview_client.get_entity(qualifiedName=qualifiedName, typeName=type)

        # get the corresponding features belongs to a certain project.
        # note that we need to use "eq" to filter exactly this qualified name
        # see syntax here: https://docs.microsoft.com/en-us/rest/api/purview/catalogdataplane/discovery/query#discovery_query_andornested
        query_filter = {
            "attributeName": "qualifiedName",
            "operator": "eq",
            "attributeValue": qualifiedName
        }
        result = self.purview_client.discovery.query(keywords = None, filter=query_filter)
        entities = result['value']
        # There should be exactly one result, but we don't enforce the check here
        for entity in entities:
            if entity.get('qualifiedName') == qualifiedName:
                return entity.get('id')

    def search_features(self, searchTerm):
        """
        Search the registry for the given query term
        For a ride hailing company few examples could be - "taxi", "passenger", "fare" etc.
        It's a keyword search on the registry metadata
        """        
        entities = self.purview_client.discovery.search_entities(searchTerm)
        return entities
    
    def _list_registered_entities_with_details(self, project_name: str, entity_type: Union[str, List[str]] = None, limit=1000, starting_offset=0,) -> List[Dict]:
        """
        List all the already registered entities. entity_type should be one of: SOURCE, DERIVED_FEATURE, ANCHOR, ANCHOR_FEATURE, FEATHR_PROJECT, or a list of those values
        limit: a maximum 1000 will be enforced at the underlying API
        
        returns a list of the result entities.
        """
        entity_type_list = [entity_type] if isinstance(
            entity_type, str) else entity_type

        for i in entity_type_list:
            if i not in {TYPEDEF_SOURCE, TYPEDEF_DERIVED_FEATURE, TYPEDEF_ANCHOR, TYPEDEF_ANCHOR_FEATURE, TYPEDEF_FEATHR_PROJECT}:
                raise RuntimeError(
                    f'only SOURCE, DERIVED_FEATURE, ANCHOR, ANCHOR_FEATURE, FEATHR_PROJECT are supported when listing the registered entities, {entity_type} is not one of them.')

        if project_name is None:
            raise RuntimeError("You need to specify a project_name")
        # the search grammar: 
        # https://docs.microsoft.com/en-us/azure/purview/how-to-search-catalog#search-query-syntax
        # https://docs.microsoft.com/en-us/rest/api/datacatalog/data-catalog-search-syntax-reference

        # get the corresponding features belongs to a certain project.
        # note that we need to use "startswith" to filter out the features that don't belong to this project.
        # see syntax here: https://docs.microsoft.com/en-us/rest/api/purview/catalogdataplane/discovery/query#discovery_query_andornested
        # this search does the following:
        # search all the entities that start with project_name+delimiter for all the search entities
        # However, for TYPEDEF_FEATHR_PROJECT, it doesn't have delimiter in the qualifiedName
        # Hence if TYPEDEF_FEATHR_PROJECT is in the `entity_type` input, we need to search for that specifically
        # and finally "OR" the result to union them
        query_filter = {
            "or":
            [{
                "and": [{
                    # this is a list of the entity types that you want to query
                    "or": [{"entityType": e} for e in entity_type_list]
                },
                    {
                    "attributeName": "qualifiedName",
                    "operator": "startswith",
                    # use `project_name + self.registry_delimiter` to limit the search results
                    "attributeValue": project_name + self.registry_delimiter
                }]},
                # if we are querying TYPEDEF_FEATHR_PROJECT, then "union" the result by using this query
                {
                "and": [{
                    "or": [{"entityType": TYPEDEF_FEATHR_PROJECT}] if TYPEDEF_FEATHR_PROJECT in entity_type_list else None
                },
                    {
                    "attributeName": "qualifiedName",
                    "operator": "startswith",
                    "attributeValue": project_name
                }]}]
        }

        # Important properties returned includes:
        # id (the guid of the entity), name, qualifiedName, @search.score,
        # and @search.highlights
        # TODO: it might be throttled in the backend and wait for the `pyapacheatlas` to fix this
        # https://github.com/wjohnson/pyapacheatlas/issues/206
        # `pyapacheatlas` needs a bit optimization to avoid additional calls.
        result_entities = self.purview_client.discovery.search_entities(query=None, search_filter=query_filter, limit = limit)
        
        # append the guid list. Since we are using project_name + delimiter to search, all the result will be valid.
        guid_list = [entity["id"] for entity in result_entities]

        entity_res = [] if guid_list is None or len(guid_list)==0 else self.purview_client.get_entity(
            guid=guid_list)["entities"]
        return entity_res

    def get_features_from_registry(self, project_name: str) -> Tuple[List[FeatureAnchor], List[DerivedFeature]]:
        """Sync Features from registry to local workspace, given a project_name, will write project's features from registry to to user's local workspace]
        If the project is big, the return result could be huge.
        Args:
            project_name (str): project name.
        """
        project_entity = self.purview_client.get_entity(qualifiedName=project_name,typeName=TYPEDEF_FEATHR_PROJECT)['entities'][0]
        
        single_direction_process = [entity for _,entity in self.purview_client.get_entity_lineage(project_entity['guid'])['guidEntityMap'].items() \
            if entity['typeName']=='Process' and \
                (entity['attributes']['qualifiedName'].startswith(RELATION_CONTAINS) \
                    or entity['attributes']['qualifiedName'].startswith(RELATION_CONSUMES))]
        contain_relations = [x['displayText'].split(' to ') for x in single_direction_process if x['attributes']['qualifiedName'].startswith(RELATION_CONTAINS)]

        consume_relations = [x['displayText'].split(' to ') for x in single_direction_process if x['attributes']['qualifiedName'].startswith(RELATION_CONSUMES)]

        entities_under_project = [self.purview_client.get_entity(x[1])['entities'][0] for x in contain_relations if x[0]== project_entity['guid']]
        if not entities_under_project:
            # if the result is empty
            return (None, None)
        entities_dict = {x['guid']:x for x in entities_under_project}

        derived_feature_list = []

        for derived_feature_entity in [x for x in entities_under_project if x['typeName']==TYPEDEF_DERIVED_FEATURE]:
            # this will be used to generate DerivedFeature instance
            derived_feature_key_list = []
                
            for key in derived_feature_entity["attributes"]["key"]:
                derived_feature_key_list.append(TypedKey(key_column=key["keyColumn"], key_column_type=key["keyColumnType"], full_name=key["fullName"], description=key["description"], key_column_alias=key["keyColumnAlias"]))
            
            def search_for_input_feature(elem, full_relations,full_entities):
                matching_relations = [x for x in full_relations if x[0]==elem['guid']]
                target_entities = [full_entities[x[1]] for x in matching_relations]
                input_features = [x for x in target_entities if x['typeName']==TYPEDEF_ANCHOR_FEATURE \
                    or x['typeName']==TYPEDEF_DERIVED_FEATURE]
                result_features=[]
                for feature_entity in input_features:
                    key_list=[]
                    for key in feature_entity["attributes"]["key"]:
                        key_list.append(TypedKey(key_column=key["keyColumn"], key_column_type=key["keyColumnType"], full_name=key["fullName"], description=key["description"], key_column_alias=key["keyColumnAlias"]))
                    result_features.append(Feature(name=feature_entity["attributes"]["name"],
                        feature_type=self._get_feature_type_from_hocon(feature_entity["attributes"]["type"]), # stored as a hocon string, can be parsed using pyhocon
                        transform=self._get_transformation_from_dict(feature_entity["attributes"]['transformation']), #transform attributes are stored in a dict fashion , can be put in a WindowAggTransformation
                        key=key_list, # since all features inside an anchor should share the same key, pick the first one.
                        registry_tags=feature_entity["attributes"]["tags"]))
                return result_features


            all_input_features = search_for_input_feature(derived_feature_entity,consume_relations,entities_dict)
            derived_feature_list.append(DerivedFeature(name=derived_feature_entity["attributes"]["name"],
                                feature_type=self._get_feature_type_from_hocon(derived_feature_entity["attributes"]["type"]),
                                transform=self._get_transformation_from_dict(derived_feature_entity["attributes"]['transformation']),
                                key=derived_feature_key_list,
                                input_features= all_input_features,
                                registry_tags=derived_feature_entity["attributes"]["tags"]))                    
        
        # anchor_result = self.purview_client.get_entity(guid=anchor_guid)["entities"]
        anchor_list = []

        for anchor_entity in [x for x in entities_under_project if x['typeName']==TYPEDEF_ANCHOR]:
            consume_items_under_anchor = [entities_dict[x[1]] for x in consume_relations if x[0]==anchor_entity['guid']]
            source_entity = [x for x in consume_items_under_anchor if x['typeName']==TYPEDEF_SOURCE][0]

            contain_items_under_anchor = [entities_dict[x[1]] for x in contain_relations if x[0]==anchor_entity['guid']]
            anchor_features_guid = [x['guid'] for x in contain_items_under_anchor if x['typeName']==TYPEDEF_ANCHOR_FEATURE]

            anchor_list.append(FeatureAnchor(name=anchor_entity["attributes"]["name"],
                                source=HdfsSource(
                                    name=source_entity["attributes"]["name"],
                                    event_timestamp_column=source_entity["attributes"]["event_timestamp_column"],
                                    timestamp_format=source_entity["attributes"]["timestamp_format"],
                                    preprocessing=self._correct_function_indentation(source_entity["attributes"]["preprocessing"]),
                                    path=source_entity["attributes"]["path"],
                                    registry_tags=source_entity["attributes"]["tags"]
                                    ),
                                features=self._get_features_by_guid_or_entities(guid_list = anchor_features_guid, entity_list=entities_under_project),
                                registry_tags=anchor_entity["attributes"]["tags"]))

        return (anchor_list, derived_feature_list)

    def search_input_anchor_features(self,derived_guids,feature_entity_guid_mapping) ->List[str]:
        '''
        Iterate all derived features and its parent links, extract and aggregate all inputs
        '''
        stack = [x for x in derived_guids]
        result = []
        while len(stack)>0:
            current_derived_guid = stack.pop()
            current_input = feature_entity_guid_mapping[current_derived_guid]
            new_derived_features = [x["guid"] for x in current_input["attributes"]["input_derived_features"]]
            new_anchor_features = [x["guid"] for x in current_input["attributes"]["input_anchor_features"]]
            for feature_guid in new_derived_features:
                stack.append(feature_guid)
            result += new_anchor_features
            result = list(set(result))
        return result


    def _correct_function_indentation(self, user_func: str) -> str:
        """
        The function read from registry might have the wrong indentation. We need to correct those indentation.
        More specifically, we are using the inspect module to copy the function body for UDF for further submission. In that case, there will be situations like this:

        def feathr_udf1(df)
            return df

                def feathr_udf2(df)
                    return df

        For example, in Feathr test cases, there are similar patterns for `feathr_udf2`, since it's defined in another function body (the registry_test_setup method).
        This is not an ideal way of dealing with that, but we'll keep it here until we figure out a better way.
        """
        if user_func is None:
            return None
        # if user_func is a string, turn it into a list of strings so that it can be used below
        temp_udf_source_code = user_func.split('\n')
        # assuming the first line is the function name
        leading_space_num = len(temp_udf_source_code[0]) - len(temp_udf_source_code[0].lstrip())
        # strip the lines to make sure the function has the correct indentation
        udf_source_code_striped = [line[leading_space_num:] for line in temp_udf_source_code]
        # append '\n' back since it was deleted due to the previous split
        udf_source_code = [line+'\n' for line in udf_source_code_striped]
        return " ".join(udf_source_code)

    def _get_source_by_guid(self, guid, entity_list) -> Source:
        """give a entity list and the target GUID for the source entity, return a python `Source` object.
        """
        # TODO: currently return HDFS source by default. For JDBC source, it's currently implemented using HDFS Source so we should split in the future

        # there should be only one entity available
        source_entity = [x for x in entity_list if x['guid'] == guid][0]

        # if source_entity["attributes"]["path"] is INPUT_CONTEXT, it will also be assigned to this returned object
        return HdfsSource(name=source_entity["attributes"]["name"],
                event_timestamp_column=source_entity["attributes"]["event_timestamp_column"],
                timestamp_format=source_entity["attributes"]["timestamp_format"],
                preprocessing=self._correct_function_indentation(source_entity["attributes"]["preprocessing"]),
                path=source_entity["attributes"]["path"],
                registry_tags=source_entity["attributes"]["tags"]
                )



    def _get_feature_type_from_hocon(self, input_str: str) -> FeatureType:
        """Get Feature types from a HOCON config, given that we stored the feature type in a plain string.

        Args:
            input_str (str): the input string for the stored HOCON config

        Returns:
            FeatureType: feature type that can be used by Python
        """
        if not input_str:
            return None
        conf = ConfigFactory.parse_string(input_str)
        valType = conf.get_string('valType') if 'valType' in conf else conf.get_string('type.valType')
        dimensionType = conf.get_string('dimensionType') if 'dimensionType' in conf else conf.get_string('type.dimensionType')
        if dimensionType == '[INT]':
            # if it's not empty, i.e. [INT], indicating it's vectors
            if valType == 'DOUBLE':
                return DoubleVectorFeatureType()
            elif valType == 'LONG':
                return Int64VectorFeatureType()
            elif valType == 'INT':
                return Int32VectorFeatureType()
            elif valType == 'FLOAT':
                return FloatVectorFeatureType()
            else:
                logger.error("{} cannot be parsed.", valType)
        else:
            if valType == 'STRING':
                return StringFeatureType()
            elif valType == 'BYTES':
                return BytesFeatureType()
            elif valType == 'DOUBLE':
                return DoubleFeatureType()
            elif valType == 'FLOAT':
                return FloatFeatureType()
            elif valType == 'LONG':
                return Int64FeatureType()
            elif valType == 'INT':
                return Int32FeatureType()
            elif valType == 'BOOLEAN':
                return BooleanFeatureType()
            else:
                logger.error("{} cannot be parsed.", valType)

    def _get_transformation_from_dict(self, input: Dict) -> FeatureType:
        if 'transformExpr' in input:
            # it's ExpressionTransformation
            return ExpressionTransformation(input['transformExpr'])
        elif 'def_expr' in input or 'defExpr' in input: 
            agg_expr=input['def_expr'] if 'def_expr' in input else (input['defExpr'] if 'defExpr' in input else None)
            agg_func=input['agg_func']if 'agg_func' in input else (input['aggFunc'] if 'aggFunc' in input else None)
            window=input['window']if 'window' in input else None
            group_by=input['group_by']if 'group_by' in input else (input['groupBy'] if 'groupBy' in input else None)
            filter=input['filter']if 'filter' in input else None
            limit=input['limit']if 'limit' in input else None
            return WindowAggTransformation(agg_expr, agg_func, window, group_by, filter, limit)
        else:
            # no transformation function observed
            return None

    def _get_features_by_guid_or_entities(self, guid_list, entity_list) -> List[FeatureAnchor]:
        """return a python list of the features that are referenced by a list of guids. 
        If entity_list is provided, use entity_list to reconstruct those features
        This is for "anchor feature" only.
        """
        if not entity_list:
            feature_entities = self.purview_client.get_entity(guid=guid_list)["entities"]
        else:
            guid_set = set(guid_list)
            feature_entities = [x for x in entity_list if x['guid'] in guid_set]

            # raise error if we cannot find all the guid
            if len(feature_entities) != len(guid_list):
                raise RuntimeError("Number of `feature_entities` is less than provided GUID list for search. The project might be broken.")

        feature_list=[]
        
        '''
        The assumption here is , a feture could have multiple keys, and features inside an anchor should share the same set of keys. 
        So we will take any one of the feature, extract its keys , dedup them by full name, and use them to generate the key list.
        '''
        first_feature_keys = feature_entities[0]["attributes"]["key"]
        deduped_keys = dict()
        for key in first_feature_keys:
            if key['fullName'] not in deduped_keys:
                deduped_keys.setdefault(key['fullName'],key)
        key_list = [
            TypedKey(key_column=key["keyColumn"], key_column_type=key["keyColumnType"], full_name=key["fullName"], description=key["description"], key_column_alias=key["keyColumnAlias"])\
                for key in list(deduped_keys.values())
                ]
        for feature_entity in feature_entities:         
            # after get keys, put them in features
            feature_list.append(Feature(name=feature_entity["attributes"]["name"],
                    feature_type=self._get_feature_type_from_hocon(feature_entity["attributes"]["type"]), # stored as a hocon string, can be parsed using pyhocon
                    transform=self._get_transformation_from_dict(feature_entity["attributes"]['transformation']), #transform attributes are stored in a dict fashion , can be put in a WindowAggTransformation
                    key=key_list, 
                    registry_tags=feature_entity["attributes"]["tags"],
            
            ))
        return feature_list 



from abc import abstractmethod
import copy
import json
from typing import Dict, List, Optional
from jinja2 import Template
from feathr.definition.feathrconfig import HoconConvertible


class Sink(HoconConvertible):
    """A data sink.
    """
    
    @abstractmethod
    def support_offline(self) -> bool:
        pass
    
    @abstractmethod
    def support_online(self) -> bool:
        pass
    
    @abstractmethod
    def to_argument(self):
        pass
    
    def __str__(self) -> str:
        return "DUMMY"

class MonitoringSqlSink(Sink):
    """SQL-based sink that stores feature monitoring results.

    Attributes:
        table_name: output table name
    """
    def __init__(self, table_name: str) -> None:
        self.table_name = table_name

    def to_feature_config(self) -> str:
        """Produce the config used in feature monitoring"""
        tm = Template("""  
            {
                name: MONITORING
                params: {
                    table_name: "{{source.table_name}}"
                }
            }
        """)
        msg = tm.render(source=self)
        return msg

    def support_offline(self) -> bool:
        return False
    
    def support_online(self) -> bool:
        return True
    
    def to_argument(self):
        raise TypeError("MonitoringSqlSink cannot be used as output argument")

class RedisSink(Sink):
    """Redis-based sink use to store online feature data, can be used in batch job or streaming job.

    Attributes:
        table_name: output table name
        streaming: whether it is used in streaming mode
        streamingTimeoutMs: maximum running time for streaming mode. It is not used in batch mode.
    """
    def __init__(self, table_name: str, streaming: bool=False, streamingTimeoutMs: Optional[int]=None) -> None:
        self.table_name = table_name
        self.streaming = streaming
        self.streamingTimeoutMs = streamingTimeoutMs

    def to_feature_config(self) -> str:
        """Produce the config used in feature materialization"""
        tm = Template("""  
            {
                name: REDIS
                params: {
                    table_name: "{{source.table_name}}"
                    {% if source.streaming %}
                    streaming: true
                    {% endif %}
                    {% if source.streamingTimeoutMs %}
                    timeoutMs: {{source.streamingTimeoutMs}}
                    {% endif %}
                    {% if source.aggregation_features %}
                    features: [{{','.join(source.aggregation_features)}}]
                    {% endif %}
                }
            }
        """)
        msg = tm.render(source=self)
        return msg

    def support_offline(self) -> bool:
        return False
    
    def support_online(self) -> bool:
        return True
    
    def to_argument(self):
        raise TypeError("RedisSink cannot be used as output argument")


class HdfsSink(Sink):
    """Offline Hadoop HDFS-compatible(HDFS, delta lake, Azure blog storage etc) sink that is used to store feature data.
    The result is in AVRO format. 

    Incremental aggregation is enabled by default when using HdfsSink. Use incremental aggregation will significantly expedite the WindowAggTransformation feature calculation. 
    For example, aggregation sum of a feature F within a 180-day window at day T can be expressed as: F(T) = F(T - 1)+DirectAgg(T-1)-DirectAgg(T - 181). 
    Once a SNAPSHOT of the first day is generated, the calculation for the following days can leverage it.  

    Attributes:
        output_path: output path
        store_name: the folder name under the base "path". Used especially for the current dataset to support 'Incremental' aggregation. 
        
    """
    def __init__(self, output_path: str, store_name: Optional[str]="df0") -> None:
        self.output_path = output_path
        self.store_name = store_name
    # Sample generated HOCON config:
    # operational: {
    #     name: testFeatureGen
    #     endTime: 2019-05-01
    #     endTimeFormat: "yyyy-MM-dd"
    #     resolution: DAILY
    #     enableIncremental = true
    #     output:[
    #         {
    #             name: HDFS
    #             outputFormat: RAW_DATA
    #             params: {
    #                 path: "/user/featureGen/hdfsResult/"
    #                 features: [mockdata_a_ct_gen, mockdata_a_sample_gen]
    #                 storeName: "yyyy/MM/dd"
    #             }
    #         }
    #     ]
    # }
    # features: [mockdata_a_ct_gen, mockdata_a_sample_gen]
    def to_feature_config(self) -> str:
        """Produce the config used in feature materialization"""
        tm = Template("""  
            {
                name: HDFS
                outputFormat: RAW_DATA
                params: {
                    path: "{{sink.output_path}}"
                    {% if sink.aggregation_features %}
                    features: [{{','.join(sink.aggregation_features)}}]
                    {% endif %}
                    {% if sink.store_name %}
                    storeName: "{{sink.store_name}}"
                    {% endif %}
                }
            }
        """)
        hocon_config = tm.render(sink=self)
        return hocon_config

    def support_offline(self) -> bool:
        return True
    
    def support_online(self) -> bool:
        return True
    
    def to_argument(self):
        return self.output_path

class JdbcSink(Sink):
    def __init__(self, name: str, url: str, dbtable: str, auth: Optional[str] = None) -> None:
        self.name = name
        self.url = url
        self.dbtable = dbtable
        if auth is not None:
            self.auth = auth.upper()
            if self.auth not in ["USERPASS", "TOKEN"]:
                raise ValueError(
                    "auth must be None or one of following values: ['userpass', 'token']")

    def get_required_properties(self):
        if not hasattr(self, "auth"):
            return []
        if self.auth == "USERPASS":
            return ["%s_USER" % self.name.upper(), "%s_PASSWORD" % self.name.upper()]
        elif self.auth == "TOKEN":
            return ["%s_TOKEN" % self.name.upper()]

    def support_offline(self) -> bool:
        return True
    
    def support_online(self) -> bool:
        return True
    
    def to_feature_config(self) -> str:
        """Produce the config used in feature materialization"""
        tm = Template("""  
            {
                name: HDFS
                params: {
                    type: "jdbc"
                    url: "{{sink.url}}"
                    dbtable: "{{sink.dbtable}}"
                    {% if sink.auth is defined %}
                        {% if sink.auth == "USERPASS" %}
                    user: "${{ "{" }}{{sink.name}}_USER{{ "}" }}"
                    password: "${{ "{" }}{{sink.name}}_PASSWORD{{ "}" }}"
                        {% else %}
                    token: "${{ "{" }}{{sink.name}}_TOKEN{{ "}" }}"
                        {% endif %}
                    {% endif %}
                }
            }
        """)
        sink = copy.copy(self)
        sink.name = self.name.upper()
        hocon_config = tm.render(sink=sink)
        return hocon_config

    def to_argument(self):
        d = {
            "type": "jdbc",
            "url": self.url,
        }
        if hasattr(self, "dbtable"):
            d["dbtable"] = self.dbtable
        if hasattr(self, "auth"):
            if self.auth == "USERPASS":
                d["user"] = "${" + self.name.upper() + "_USER}"
                d["password"] = "${" + self.name.upper() + "_PASSWORD}"
            elif self.auth == "TOKEN":
                d["useToken"] = True
                d["token"] = "${" + self.name.upper() + "_TOKEN}"
        else:
            d["anonymous"] = True
        return json.dumps(d)
    
class GenericSink(Sink):
    """
    This class is corresponding to 'GenericLocation' in Feathr core, but only be used as Sink.
    The class is not meant to be used by user directly, user should use its subclasses like `CosmosDbSink`
    """
    def __init__(self, format: str, mode: Optional[str] = None, options: Dict[str, str] = {}) -> None:
        self.format = format
        self.mode = mode
        self.options = dict([(o.replace(".", "__"), options[o]) for o in options])
    
    def to_feature_config(self) -> str:
        ret = {
            "name": "HDFS",
            "params": self._to_dict()
        }
        return json.dumps(ret, indent=4)
    
    def _to_dict(self) -> Dict[str, str]:
        ret = self.options.copy()
        ret["type"] = "generic"
        ret["format"] = self.format
        if self.mode:
            ret["mode"] = self.mode
        return ret        
    
    def get_required_properties(self):
        ret = []
        for option in self.options:
            start = option.find("${")
            if start >= 0:
                end = option[start:].find("}")
                if end >= 0:
                    ret.append(option[start+2:start+end])
        return ret

    def to_argument(self):
        """
        One-line JSON string, used by job submitter
        """
        return json.dumps(self._to_dict())
    
class CosmosDbSink(GenericSink):
    """
    CosmosDbSink is a sink that is used to store online feature data in CosmosDB.
    Even it's possible, but we shouldn't use it as offline store as CosmosDb requires records to have unique keys, why offline feature job cannot generate unique keys.
    """
    def __init__(self, name: str, endpoint: str, database: str, container: str):        
        super().__init__(format = "cosmos.oltp", mode="APPEND", options={
            "spark.cosmos.accountEndpoint": endpoint,
            'spark.cosmos.accountKey': "${%s_KEY}" % name.upper(),
            "spark.cosmos.database": database,
            "spark.cosmos.container": container
        })
        self.name = name
        self.endpoint = endpoint
        self.database = database
        self.container = container
        
    def support_offline(self) -> bool:
        return False
    
    def support_online(self) -> bool:
        return True
    
    def get_required_properties(self) -> List[str]:
        return [self.name.upper() + "_KEY"]

class ElasticSearchSink(GenericSink):
    """
    Use ElasticSearch as the data sink.
    """
    def __init__(self,
                 name: str,
                 host: str,
                 index: str,
                 ssl: bool = True,
                 auth: bool = True,
                 mode = 'OVERWRITE'):
        """
        name: The name of the sink.
        host: ElasticSearch node, can be `hostname` or `hostname:port`, default port is 9200.
        index: The index to write the data.
        ssl: Set to `True` to enable SSL.
        auth: Set to `True` to enable authentication, you need to provide username/password from environment or KeyVault.
        mode: Spark mode, check official doc for more details.
        """
        self.auth = auth
        options = {
            'es.nodes': host,
            'es.ssl': str(ssl).lower(),
            'es.resource': index,
        }
        if auth:
            """
            Currently only BasicAuth is supported.
            ElasticSearch Spark connector also supports PKI auth but that needs to setup keystore on each driver node,
            which seems to be too complicated for managed Spark cluster.
            """
            options["es.net.http.auth.user"] = "${%s_USER}" % name.upper(),
            options["es.net.http.auth.pass"] = "${%s_PASSWORD}" % name.upper(),
        super().__init__(name,
                         format='org.elasticsearch.spark.sql',
                         mode=mode,
                         options=options)

    def support_offline(self) -> bool:
        """
        CAUTION: Using ES as offline store is possible, but the FeatureJoinJob output doesn't have a key column, you need to make sure
        the output dataset is accessible in other ways, like full-text search or time-series with a timestamp field.
        """
        return True
    
    def support_online(self) -> bool:
        return True
    
    def get_required_properties(self) -> List[str]:
        if self.auth:
            return [self.name.upper() + "_USER", self.name.upper() + "_PASSWORD"]
        return []

class AerospikeSink(GenericSink):
    def __init__(self,name:str,seedhost:str,port:int,namespace:str,setname:str):
        super().__init__(format="aerospike",mode="APPEND",options = {
            "aerospike.seedhost":seedhost,
            "aerospike.port":str(port),
            "aerospike.namespace":namespace,
            "aerospike.user":"${%s_USER}" % name.upper(),
            "aerospike.password":"${%s_PASSWORD}" % name.upper(),
            "aerospike.set":setname
        })
        self.name = name

    def support_offline(self) -> bool:
        return False
    
    def support_online(self) -> bool:
        return True
    
    def get_required_properties(self) -> List[str]:
        return [self.name.upper() + "_USER", self.name.upper() + "_PASSWORD"]



import os
from pathlib import Path
import pytest
from feathr import MaterializationSettings, RedisSink
from test_fixture import kafka_test_setup
from test_utils.constants import Constants

@pytest.mark.skipif(os.environ.get('SPARK_CONFIG__SPARK_CLUSTER') != "azure_synapse",
                    reason="skip for databricks, as it cannot stop streaming job automatically for now.")
def test_feathr_kafa_streaming_features():
    """
    Test FeathrClient() materialize_features can ingest streaming feature correctly
    """
    test_workspace_dir = Path(__file__).parent.resolve() / "test_user_workspace"

    client = kafka_test_setup(os.path.join(test_workspace_dir, "feathr_config.yaml"))
    redisSink = RedisSink(table_name="kafkaSampleDemoFeature", streaming=True, streamingTimeoutMs=10000)
    settings = MaterializationSettings(name="kafkaSampleDemo",
                                   sinks=[redisSink],
                                   feature_names=['f_modified_streaming_count']
                                   )
    client.materialize_features(settings, allow_materialize_non_agg_feature=True)
    client.wait_job_to_finish(timeout_sec=Constants.SPARK_JOB_TIMEOUT_SECONDS)




from pyspark.sql import SparkSession, DataFrame, SQLContext
import sys
from pyspark.sql.functions import *

# This is executed in Spark driver
# The logger doesn't work in Pyspark so we just use print
print("Feathr Pyspark job started.")
spark = SparkSession.builder.appName('FeathrPyspark').getOrCreate()


def to_java_string_array(arr):
    """Convert a Python string list to a Java String array.
    """
    jarr = spark._sc._gateway.new_array(spark._sc._jvm.java.lang.String, len(arr))
    for i in range(len(arr)):
        jarr[i] = arr[i]
    return jarr


def submit_spark_job(feature_names_funcs):
    """Submit the Pyspark job to the cluster. This should be used when there is Python UDF preprocessing for sources.
    It loads the source DataFrame from Scala spark. Then preprocess the DataFrame with Python UDF in Pyspark. Later,
    the real Scala FeatureJoinJob or FeatureGenJob is executed with preprocessed DataFrames instead of the original
    source DataFrames.

        Args:
            feature_names_funcs: Map of feature names concatenated to preprocessing UDF function.
            For example {"f1,f2": df1, "f3,f4": df2} (the feature names in the key will be sorted)
    """
    # Prepare job parameters
    # sys.argv has all the arguments passed by submit job.
    # In pyspark job, the first param is the python file.
    # For example: ['pyspark_client.py', '--join-config', 'abfss://...', ...]
    has_gen_config = False
    has_join_config = False
    if '--generation-config' in sys.argv:
        has_gen_config = True
    if '--join-config' in sys.argv:
        has_join_config = True

    py4j_feature_job = None
    if has_gen_config and has_join_config:
        raise RuntimeError("Both FeatureGenConfig and FeatureJoinConfig are provided. "
                           "Only one of them should be provided.")
    elif has_gen_config:
        py4j_feature_job = spark._jvm.com.linkedin.feathr.offline.job.FeatureGenJob
        print("FeatureGenConfig is provided. Executing FeatureGenJob.")
    elif has_join_config:
        py4j_feature_job = spark._jvm.com.linkedin.feathr.offline.job.FeatureJoinJob
        print("FeatureJoinConfig is provided. Executing FeatureJoinJob.")
    else:
        raise RuntimeError("None of FeatureGenConfig and FeatureJoinConfig are provided. "
                           "One of them should be provided.")
    job_param_java_array = to_java_string_array(sys.argv)

    print("submit_spark_job: feature_names_funcs: ")
    print(feature_names_funcs)
    print("set(feature_names_funcs.keys()): ")
    print(set(feature_names_funcs.keys()))

    print("submit_spark_job: Load DataFrame from Scala engine.")

    dataframeFromSpark = py4j_feature_job.loadSourceDataframe(job_param_java_array, set(feature_names_funcs.keys())) # TODO: Add data handler support here
    print("Submit_spark_job: dataframeFromSpark: ")
    print(dataframeFromSpark)

    # per comment https://stackoverflow.com/a/54738984, use explicit way to initialize SQLContext
    # Otherwise it might fail when calling `DataFrame.collect()` or other APIs that's related with SQLContext
    # do not use `sql_ctx = SQLContext(spark)`
    sql_ctx = SQLContext(sparkContext=spark.sparkContext, sparkSession=spark)
    new_preprocessed_df_map = {}
    for feature_names, scala_dataframe in dataframeFromSpark.items():
        # Need to convert java DataFrame into python DataFrame
        py_df = DataFrame(scala_dataframe, sql_ctx)
        # Preprocess the DataFrame via UDF
        user_func = feature_names_funcs[feature_names]
        preprocessed_udf = user_func(py_df)
        new_preprocessed_df_map[feature_names] = preprocessed_udf._jdf

    print("submit_spark_job: running Feature job with preprocessed DataFrames:")
    print("Preprocessed DataFrames are: ")
    print(new_preprocessed_df_map)

    py4j_feature_job.mainWithPreprocessedDataFrame(job_param_java_array, new_preprocessed_df_map)
    return None




from feathrcli.cli import init
from click.testing import CliRunner
import os
import glob


def test_workspace_creation():
    """
    Test CLI init() is working properly.
    """
    runner = CliRunner()
    with runner.isolated_filesystem():

        result = runner.invoke(init, [])

        assert result.exit_code == 0
        assert os.path.isdir("./feathr_user_workspace")

        total_yaml_files = glob.glob('./feathr_user_workspace/*.yaml', recursive=True)
        # we should have exact 1 yaml file
        assert len(total_yaml_files) == 1
        # result = runner.invoke(init, [])

        test_folder_name = 'test_folder'
        result = runner.invoke(init, ['--name', test_folder_name])

        assert result.exit_code == 0

        total_yaml_files = glob.glob(os.path.join(test_folder_name, '*.yaml'), recursive=True)
        # we should have exact 1 yaml file
        assert len(total_yaml_files) == 1

        result = runner.invoke(init, ["--name", test_folder_name])

        assert result.exit_code == 2
        # use output for test for now
        expected_out = f'Feathr workspace ({test_folder_name}) already exist. Please use a new folder name.\n'
        assert expected_out in result.output



from feathr.anchor import FeatureAnchor
from feathr.source import HdfsSource
from feathr.feature import Feature
from feathr.dtype import BOOLEAN, FLOAT, ValueType
from feathr.transformation import WindowAggTransformation
from feathr.typed_key import TypedKey

batch_source = HdfsSource(name="nycTaxiBatchSource",
                          path="wasbs://public@azurefeathrstorage.blob.core.windows.net/sample_data/green_tripdata_2020-04.csv",
                          event_timestamp_column="lpep_dropoff_datetime",
                          timestamp_format="yyyy-MM-dd HH:mm:ss")

location_id = TypedKey(key_column="DOLocationID",
                       key_column_type=ValueType.INT32,
                       description="location id in NYC",
                       full_name="nyc_taxi.location_id")
agg_features = [Feature(name="f_location_avg_fare",
                        key=location_id,
                        feature_type=FLOAT,
                        transform=WindowAggTransformation(agg_expr="cast_float(fare_amount)",
                                                          agg_func="AVG",
                                                          window="90d")),
                Feature(name="f_location_max_fare",
                        key=location_id,
                        feature_type=FLOAT,
                        transform=WindowAggTransformation(agg_expr="cast_float(fare_amount)",
                                                          agg_func="MAX",
                                                          window="90d"))
                ]

agg_anchor = FeatureAnchor(name="aggregationFeatures",
                           source=batch_source,
                           features=agg_features)



from feathr.definition.materialization_settings import MaterializationSettings


# it's completely the same as MaterializationSettings. But we renamed it to improve usability.
# In the future, we may want to rely a separate system other than MaterializationSettings to generate stats.
class MonitoringSettings(MaterializationSettings):
    """Settings about monitoring features.
    """



from feathr.client import FeathrClient
import os
import glob
from feathr.constants import OUTPUT_FORMAT
from loguru import logger
import pandas as pd
import tempfile
from pandas.errors import EmptyDataError



def get_result_df(client: FeathrClient, format: str = None, res_url: str = None, local_folder: str = None) -> pd.DataFrame:
    """Download the job result dataset from cloud as a Pandas dataframe to make it easier for the client to read.

    format: format to read the downloaded files. Currently support `parquet`, `delta`, `avro`, and `csv`. Default to `avro` if not specified.
    res_url: output URL to download files. Note that this will not block the job so you need to make sure the job is finished and result URL contains actual data.
    local_folder: optional parameter to specify the absolute download path. if the user does not provide this, function will create a temporary directory and delete it after reading the dataframe.
    """
    # use a result url if it's provided by the user, otherwise use the one provided by the job 
    res_url: str = res_url or client.get_job_result_uri(block=True, timeout_sec=1200)
    if res_url is None:
        raise RuntimeError("res_url is None. Please make sure either you provide a res_url or make sure the job finished in FeathrClient has a valid result URI.")

    # use user provided format, if there isn't one, then otherwise use the one provided by the job; 
    # if none of them is available, "avro" is the default format.
    format: str = format or client.get_job_tags().get(OUTPUT_FORMAT, "")
    if format is None or format == "":
        format = "avro"

    # if local_folder params is not provided then create a temporary folder
    if local_folder is not None:
        local_dir_path = local_folder
    else:
        tmp_dir = tempfile.TemporaryDirectory()
        local_dir_path = tmp_dir.name
        
    client.feathr_spark_launcher.download_result(result_path=res_url, local_folder=local_dir_path)
    dataframe_list = []
    # by default the result are in avro format
    if format.casefold()=="parquet":
        files =  glob.glob(os.path.join(local_dir_path, '*.parquet'))
        from pyarrow.parquet import ParquetDataset
        ds = ParquetDataset(files)
        result_df = ds.read().to_pandas()
    elif format.casefold()=="delta":
        from deltalake import DeltaTable
        delta = DeltaTable(local_dir_path)
        if not client.spark_runtime == 'azure_synapse':
            # don't detect for synapse result with Delta as there's a problem with underlying system
            # Issues are tracked here: https://github.com/delta-io/delta-rs/issues/582
            result_df = delta.to_pyarrow_table().to_pandas()
        else:
            logger.info("Please use Azure Synapse to read the result in the Azure Synapse cluster. Reading local results is not supported for Azure Synapse. Empty DataFrame is returned.")
            result_df = pd.DataFrame()
    elif format.casefold()=="avro":
        import pandavro as pdx
        for file in glob.glob(os.path.join(local_dir_path, '*.avro')):
            dataframe_list.append(pdx.read_avro(file))
        result_df = pd.concat(dataframe_list, axis=0)
    elif format.casefold()=="csv":
        for file in glob.glob(os.path.join(local_dir_path, '*.csv')):
            try:
                df = pd.read_csv(file, index_col=None, header=None)
            except EmptyDataError:
                # in case there are empty files
                df = pd.DataFrame()
            dataframe_list.append(df)
        result_df = pd.concat(dataframe_list, axis=0)
        # Reset index to avoid duplicated indices
        result_df.reset_index(drop=True)
    else:
        raise RuntimeError(f"{format} is currently not supported in get_result_df. Currently only parquet, delta, avro, and csv are supported, please consider writing a customized function to read the result.")

    
    if local_folder is None:
        tmp_dir.cleanup()
    return result_df


from feathr.anchor import FeatureAnchor
from feathr.feature import Feature
from feathr.dtype import BOOLEAN, INT32, FLOAT, ValueType
from feathr.feature_derivations import DerivedFeature
from feathr.source import INPUT_CONTEXT

f_trip_distance = Feature(name="f_trip_distance", feature_type=FLOAT, transform="trip_distance")
f_trip_time_duration = Feature(name="f_trip_time_duration",
            feature_type=INT32,
            transform="(to_unix_timestamp(lpep_dropoff_datetime) - to_unix_timestamp(lpep_pickup_datetime))/60")

features = [
    f_trip_distance,
    f_trip_time_duration,
    Feature(name="f_is_long_trip_distance",
            feature_type=BOOLEAN,
            transform="cast_float(trip_distance)>30"),
    Feature(name="f_day_of_week",
            feature_type=INT32,
            transform="dayofweek(lpep_dropoff_datetime)"),
  ]

request_anchor = FeatureAnchor(name="request_features",
                source=INPUT_CONTEXT,
                features=features)


f_trip_time_distance = DerivedFeature(name="f_trip_time_distance",
                feature_type=FLOAT,
                input_features=[f_trip_distance, f_trip_time_duration],
                transform="f_trip_distance * f_trip_time_duration")

f_trip_time_rounded = DerivedFeature(name="f_trip_time_rounded",
                feature_type=INT32,
                input_features=[f_trip_time_duration],
                transform="f_trip_time_duration % 10")
